import networkx as nx
import random
EDGE_TEMPLATES = [
  "{}_{}_OP"]
NODE_TEMPLATES = [
  "{}_activation",
  "{}_normalisation",
  "{}_stride",
  "{}_channel_ratio",
  "{}_combine"]

def random_strides(ops,count,data,stride, channel_ratio):
  s = [ x for x in ops if "stride" in x]
  while count > 0:
    o = random.choice(s)
    ops[o] *= stride
    ops["{}_channel_ratio".format(o.split("_")[0])] *=channel_ratio # FIX THIS ALSO 
    count -= 1

  """
  for i in s:
    ops[i] = 2 ** (ops[i] - 1 )
    ops["{}_channel_ratio".format(i.split("_")[0])] = 2 ** (ops["{}_channel_ratio".format(i.split("_")[0])] - 1 )
    
  """
  return ops


def generate_skip(graph):
  ops = {}
  nodes = graph.nodes()
  edges = graph.edges()
  for template in NODE_TEMPLATES:
    for node in nodes:
        if "stride" in template or "channel" in template:
          ops[template.format(node)] = 1

        else:
          ops[template.format(node)] = "none"
  for template in EDGE_TEMPLATES:
    for edge in edges:
      ops[template.format(*edge)] = "skip_connect"
  return ops

def generate_op_names(graph):
  ops = {}
  nodes = graph.nodes()
  edges = graph.edges()
  for template in NODE_TEMPLATES:
    for node in nodes:
        if "stride" or "channel" in template:
          ops[template.format(node)] = 1
        else:
          ops[template.format(node)] = "none"
  for template in EDGE_TEMPLATES:
    for edge in edges:
      ops[template.format(*edge)] = "none"
  return ops

def random_ops_unweighted(ops, data):
  for i in ops:
    if "OP" in i:
      ops[i] = random.choice(data["EDGE_OPERATIONS"])
  return ops

def random_activation_unweighted(ops,data):
  for i in ops:
    if "activation" in i:
      ops[i] = random.choice(data["ACTIVATION_FUNCTIONS"])
  return ops

def random_combine_unweighted(ops,data):
  for i in ops:
    if "combine" in i:
      ops[i] = random.choice(data["COMBINE_OPERATIONS"])
  return ops

def random_normalisation_unweighted(ops,data):
  for i in ops:
    if "normalisation" in i:
      ops[i] = random.choice(data["NORMALISATION_FUNCTIONS"])
  return ops

def graph_joiner(old,new):
  #FLATTEN THE GRAPH TO GET HIGHEST NODE (USING NETWORKX)
  g = nx.DiGraph()
  g.add_edges_from(old)
  _max = 0
  for e in g.nodes():
    if type(e) == int:
      if int(e) > int(_max):
        _max = e
  #ADD MAX TO ALL VALUES IN NEW SET
  for i in new:
    if i[0] == "S":
      old.append((i[0] , i[1] + _max))
    else:
      old.append((i[0]+_max , i[1] + _max))


  return old

def op_joiner(old,new):
  _max = 0
  for n in old:
    for i in n.split("_"):
      if i.isdigit():
        if int(i) > int(_max):
          _max = i

  for n in new:
    split_key = n.split("_")
    for e,i in enumerate(split_key):
      if i.isdigit():
        split_key[e] = str(int(i) + int(_max))
    old["_".join(split_key)] = new[n]
  return old 



import torch
import torch.nn as nn
import torch.nn.functional as F


def correct_sizes(sizes):
    corrected_sizes = [s if s % 2 != 0 else s - 1 for s in sizes]
    return corrected_sizes


def pass_through(X):
    return X


class Inception(nn.Module):
    def __init__(self, in_channels, n_filters, kernel_sizes=[9, 19, 39], bottleneck_channels=32, activation=nn.ReLU(), return_indices=False):
        """
        : param in_channels             Number of input channels (input features)
        : param n_filters               Number of filters per convolution layer => out_channels = 4*n_filters
        : param kernel_sizes            List of kernel sizes for each convolution.
                                        Each kernel size must be odd number that meets -> "kernel_size % 2 !=0".
                                        This is nessesery because of padding size.
                                        For correction of kernel_sizes use function "correct_sizes". 
        : param bottleneck_channels     Number of output channels in bottleneck. 
                                        Bottleneck wont be used if nuber of in_channels is equal to 1.
        : param activation              Activation function for output tensor (nn.ReLU()). 
        : param return_indices          Indices are needed only if we want to create decoder with InceptionTranspose with MaxUnpool1d. 
        """
        super(Inception, self).__init__()
        self.return_indices=return_indices
        if in_channels > 1:
            self.bottleneck = nn.Conv1d(
                                in_channels=in_channels, 
                                out_channels=bottleneck_channels, 
                                kernel_size=1, 
                                stride=1, 
                                bias=False
                                )
        else:
            self.bottleneck = pass_through
            bottleneck_channels = 1

        self.conv_from_bottleneck_1 = nn.Conv1d(
                                        in_channels=bottleneck_channels, 
                                        out_channels=n_filters, 
                                        kernel_size=kernel_sizes[0], 
                                        stride=1, 
                                        padding=kernel_sizes[0]//2, 
                                        bias=False
                                        )
        self.conv_from_bottleneck_2 = nn.Conv1d(
                                        in_channels=bottleneck_channels, 
                                        out_channels=n_filters, 
                                        kernel_size=kernel_sizes[1], 
                                        stride=1, 
                                        padding=kernel_sizes[1]//2, 
                                        bias=False
                                        )
        self.conv_from_bottleneck_3 = nn.Conv1d(
                                        in_channels=bottleneck_channels, 
                                        out_channels=n_filters, 
                                        kernel_size=kernel_sizes[2], 
                                        stride=1, 
                                        padding=kernel_sizes[2]//2, 
                                        bias=False
                                        )
        self.max_pool = nn.MaxPool1d(kernel_size=3, stride=1, padding=1, return_indices=return_indices)
        self.conv_from_maxpool = nn.Conv1d(
                                    in_channels=in_channels, 
                                    out_channels=n_filters, 
                                    kernel_size=1, 
                                    stride=1,
                                    padding=0, 
                                    bias=False
                                    )
        self.batch_norm = nn.BatchNorm1d(num_features=4*n_filters)
        self.activation = activation

    def forward(self, X):
        # step 1
        Z_bottleneck = self.bottleneck(X)
        if self.return_indices:
            Z_maxpool, indices = self.max_pool(X)
        else:
            Z_maxpool = self.max_pool(X)
        # step 2
        Z1 = self.conv_from_bottleneck_1(Z_bottleneck)
        Z2 = self.conv_from_bottleneck_2(Z_bottleneck)
        Z3 = self.conv_from_bottleneck_3(Z_bottleneck)
        Z4 = self.conv_from_maxpool(Z_maxpool)
        # step 3 
        Z = torch.cat([Z1, Z2, Z3, Z4], axis=1)
        Z = self.activation(self.batch_norm(Z))
        if self.return_indices:
            return Z, indices
        else:
            return Z


class InceptionBlock(nn.Module):
    def __init__(self, in_channels, n_filters=32, kernel_sizes=[9,19,39], bottleneck_channels=32, use_residual=True, activation=nn.ReLU(), return_indices=False):
        super(InceptionBlock, self).__init__()
        self.use_residual = use_residual
        self.return_indices = return_indices
        self.activation = activation
        self.inception_1 = Inception(
                            in_channels=in_channels,
                            n_filters=n_filters,
                            kernel_sizes=kernel_sizes,
                            bottleneck_channels=bottleneck_channels,
                            activation=activation,
                            return_indices=return_indices
                            )
        self.inception_2 = Inception(
                            in_channels=4*n_filters,
                            n_filters=n_filters,
                            kernel_sizes=kernel_sizes,
                            bottleneck_channels=bottleneck_channels,
                            activation=activation,
                            return_indices=return_indices
                            )
        self.inception_3 = Inception(
                            in_channels=4*n_filters,
                            n_filters=n_filters,
                            kernel_sizes=kernel_sizes,
                            bottleneck_channels=bottleneck_channels,
                            activation=activation,
                            return_indices=return_indices
                            )   
        if self.use_residual:
            self.residual = nn.Sequential(
                                nn.Conv1d(
                                    in_channels=in_channels, 
                                    out_channels=4*n_filters, 
                                    kernel_size=1,
                                    stride=1,
                                    padding=0
                                    ),
                                nn.BatchNorm1d(
                                    num_features=4*n_filters
                                    )
                                )

    def forward(self, X):
        if self.return_indices:
            Z, i1 = self.inception_1(X)
            Z, i2 = self.inception_2(Z)
            Z, i3 = self.inception_3(Z)
        else:
            Z = self.inception_1(X)
            Z = self.inception_2(Z)
            Z = self.inception_3(Z)
        if self.use_residual:
            Z = Z + self.residual(X)
            Z = self.activation(Z)
        if self.return_indices:
            return Z,[i1, i2, i3]
        else:
            return Z



class InceptionTranspose(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_sizes=[9, 19, 39], bottleneck_channels=32, activation=nn.ReLU()):
        """
        : param in_channels             Number of input channels (input features)
        : param n_filters               Number of filters per convolution layer => out_channels = 4*n_filters
        : param kernel_sizes            List of kernel sizes for each convolution.
                                        Each kernel size must be odd number that meets -> "kernel_size % 2 !=0".
                                        This is nessesery because of padding size.
                                        For correction of kernel_sizes use function "correct_sizes". 
        : param bottleneck_channels     Number of output channels in bottleneck. 
                                        Bottleneck wont be used if nuber of in_channels is equal to 1.
        : param activation              Activation function for output tensor (nn.ReLU()). 
        """
        super(InceptionTranspose, self).__init__()
        self.activation = activation
        self.conv_to_bottleneck_1 = nn.ConvTranspose1d(
                                        in_channels=in_channels, 
                                        out_channels=bottleneck_channels, 
                                        kernel_size=kernel_sizes[0], 
                                        stride=1, 
                                        padding=kernel_sizes[0]//2, 
                                        bias=False
                                        )
        self.conv_to_bottleneck_2 = nn.ConvTranspose1d(
                                        in_channels=in_channels, 
                                        out_channels=bottleneck_channels, 
                                        kernel_size=kernel_sizes[1], 
                                        stride=1, 
                                        padding=kernel_sizes[1]//2, 
                                        bias=False
                                        )
        self.conv_to_bottleneck_3 = nn.ConvTranspose1d(
                                        in_channels=in_channels, 
                                        out_channels=bottleneck_channels, 
                                        kernel_size=kernel_sizes[2], 
                                        stride=1, 
                                        padding=kernel_sizes[2]//2, 
                                        bias=False
                                        )
        self.conv_to_maxpool = nn.Conv1d(
                                    in_channels=in_channels, 
                                    out_channels=out_channels, 
                                    kernel_size=1, 
                                    stride=1,
                                    padding=0, 
                                    bias=False
                                    )
        self.max_unpool = nn.MaxUnpool1d(kernel_size=3, stride=1, padding=1)
        self.bottleneck = nn.Conv1d(
                                in_channels=3*bottleneck_channels, 
                                out_channels=out_channels, 
                                kernel_size=1, 
                                stride=1, 
                                bias=False
                                )
        self.batch_norm = nn.BatchNorm1d(num_features=out_channels)

        def forward(self, X, indices):
            Z1 = self.conv_to_bottleneck_1(X)
            Z2 = self.conv_to_bottleneck_2(X)
            Z3 = self.conv_to_bottleneck_3(X)
            Z4 = self.conv_to_maxpool(X)

            Z = torch.cat([Z1, Z2, Z3], axis=1)
            MUP = self.max_unpool(Z4, indices)
            BN = self.bottleneck(Z)
            # another possibility insted of sum BN and MUP is adding 2nd bottleneck transposed convolution
            
            return self.activation(self.batch_norm(BN + MUP))


class InceptionTransposeBlock(nn.Module):
    def __init__(self, in_channels, out_channels=32, kernel_sizes=[9,19,39], bottleneck_channels=32, use_residual=True, activation=nn.ReLU()):
        super(InceptionTransposeBlock, self).__init__()
        self.use_residual = use_residual
        self.activation = activation
        self.inception_1 = InceptionTranspose(
                            in_channels=in_channels,
                            out_channels=in_channels,
                            kernel_sizes=kernel_sizes,
                            bottleneck_channels=bottleneck_channels,
                            activation=activation
                            )
        self.inception_2 = InceptionTranspose(
                            in_channels=in_channels,
                            out_channels=in_channels,
                            kernel_sizes=kernel_sizes,
                            bottleneck_channels=bottleneck_channels,
                            activation=activation
                            )
        self.inception_3 = InceptionTranspose(
                            in_channels=in_channels,
                            out_channels=out_channels,
                            kernel_sizes=kernel_sizes,
                            bottleneck_channels=bottleneck_channels,
                            activation=activation
                            )   
        if self.use_residual:
            self.residual = nn.Sequential(
                                nn.ConvTranspose1d(
                                    in_channels=in_channels, 
                                    out_channels=out_channels, 
                                    kernel_size=1,
                                    stride=1,
                                    padding=0
                                    ),
                                nn.BatchNorm1d(
                                    num_features=out_channels
                                    )
                                )

    def forward(self, X, indices):
        assert len(indices)==3
        Z = self.inception_1(X, indices[2])
        Z = self.inception_2(Z, indices[1])
        Z = self.inception_3(Z, indices[0])
        if self.use_residual:
            Z = Z + self.residual(X)
            Z = self.activation(Z)
        return Z




class Flatten(nn.Module):
    def __init__(self, out_features):
        super(Flatten, self).__init__()
        self.output_dim = out_features

    def forward(self, x):
        return x.view(-1, self.output_dim)
    
class Reshape(nn.Module):
    def __init__(self, out_shape):
        super(Reshape, self).__init__()
        self.out_shape = out_shape

    def forward(self, x):
        return x.view(-1, *self.out_shape)




class InceptionTime(nn.Module):
    def __init__(self,in_channels, length , num_classes ,n_filters, kernel_sizes=[9,19,39], use_residual=True, activation=nn.ReLU(), return_indices=False):
        super(InceptionTime, self).__init__()
        self.model = nn.Sequential(
                    #Reshape(out_shape=(in_channels,length)),
                    InceptionBlock(
                        in_channels=in_channels, 
                        n_filters=n_filters, 
                        kernel_sizes=kernel_sizes,
                        bottleneck_channels=n_filters,
                        use_residual=use_residual,
                        activation=activation
                    ),
                    InceptionBlock(
                        in_channels=n_filters*4, 
                        n_filters=n_filters, 
                        kernel_sizes=kernel_sizes,
                        bottleneck_channels=n_filters,
                        use_residual=use_residual,
                        activation=activation
                    ),
                    nn.AdaptiveAvgPool1d(output_size=1),
                    Flatten(out_features=n_filters*4*1),
                    nn.Linear(in_features=4*n_filters*1, out_features= num_classes)
        )

    def forward(self,x):
        return self.model(x)






class MultiResAtt(nn.Module):
    def __init__(self, input_channels, n_filters, num_classes):
        super(MultiResAtt, self).__init__()
        num_IMUs = 1
        self.initial_block = nn.Sequential(
            nn.Conv1d(input_channels, n_filters, kernel_size=3, padding=1),
            nn.BatchNorm1d(n_filters),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2)
        )

        self.residual_modules = nn.ModuleList([nn.Sequential(
            nn.Conv1d(n_filters, n_filters, kernel_size=3, padding=1),
            nn.BatchNorm1d(n_filters),
            nn.ReLU(),
            nn.Conv1d(n_filters, n_filters*2, kernel_size=3, padding=1),
            nn.BatchNorm1d(n_filters*2),
            nn.ReLU(),
            nn.Conv1d(n_filters*2, n_filters*4, kernel_size=3, padding=1),
            nn.BatchNorm1d(n_filters*4),
            nn.ReLU()
        ) for _ in range(num_IMUs)])

        self.adaptive_avg_pool = nn.AdaptiveAvgPool1d(output_size=1)

        self.bigru = nn.GRU(input_size=n_filters*4 * num_IMUs, hidden_size=n_filters*2, num_layers=2, batch_first=True, dropout=0.3, bidirectional=True)

        self.attention = nn.Sequential(
            nn.Linear(n_filters*4, n_filters*2),
            nn.Tanh(),
            nn.Linear(n_filters*2, 1)
        )

        self.fc = nn.Sequential(
            nn.Linear(n_filters*4, n_filters*2),
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Linear(n_filters*2, n_filters*2),
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Linear(n_filters*2, num_classes),
        )

    def forward(self, x):
        x = self.initial_block(x)
        
        # Apply residual modules
        residual_outputs = []
        for residual_module in self.residual_modules:
            residual_outputs.append(residual_module(x))

        # Concatenate the outputs of residual modules
        x = torch.cat(residual_outputs, dim=1)
        
        x = self.adaptive_avg_pool(x).squeeze(dim=-1)

        # Add an extra dimension for the GRU input
        x = x.unsqueeze(1)
        
        gru_out, _ = self.bigru(x)
        attention_weights = self.attention(gru_out).squeeze(dim=-1)
        attention_weights = torch.softmax(attention_weights, dim=-1).unsqueeze(-1)
        attention_out = torch.bmm(gru_out.permute(0, 2, 1), attention_weights).squeeze(dim=-1)

        output = self.fc(attention_out)
        return output


import numpy as np
from tqdm import tqdm
import time
import torch


      
def jitter(x, sigma=0.03):
    # https://arxiv.org/pdf/1706.00527.pdf
    return x + np.random.normal(loc=0., scale=sigma, size=x.shape)

def scaling(x, sigma=0.1):
    # https://arxiv.org/pdf/1706.00527.pdf
    factor = np.random.normal(loc=1., scale=sigma, size=(x.shape[0],x.shape[2]))
    return np.multiply(x, factor[:,np.newaxis,:])

def rotation(x):
    flip = np.random.choice([-1, 1], size=(x.shape[0],x.shape[2]))
    rotate_axis = np.arange(x.shape[2])
    np.random.shuffle(rotate_axis)    
    return flip[:,np.newaxis,:] * x[:,:,rotate_axis]

def permutation(x, max_segments=5, seg_mode="equal"):

    orig_steps = np.arange(x.shape[1])
    
    num_segs = np.random.randint(1, max_segments, size=(x.shape[0]))
    
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        if num_segs[i] > 1:
            if seg_mode == "random":
                split_points = np.random.choice(x.shape[1]-2, num_segs[i]-1, replace=False)
                split_points.sort()
                splits = np.split(orig_steps, split_points)
            else:
                splits = np.array_split(orig_steps, num_segs[i])
            warp = np.concatenate(np.random.permutation(splits)).ravel()
            ret[i] = pat[warp]
        else:
            ret[i] = pat
    return ret

def magnitude_warp(x, sigma=0.2, knot=4):

    from scipy.interpolate import CubicSpline
    orig_steps = np.arange(x.shape[1])
    
    random_warps = np.random.normal(loc=1.0, scale=sigma, size=(x.shape[0], knot+2, x.shape[2]))
    warp_steps = (np.ones((x.shape[2],1))*(np.linspace(0, x.shape[1]-1., num=knot+2))).T
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        warper = np.array([CubicSpline(warp_steps[:,dim], random_warps[i,:,dim])(orig_steps) for dim in range(x.shape[2])]).T
        ret[i] = pat * warper

    return ret

def time_warp(x, sigma=0.2, knot=4):
    if type(x) == torch.Tensor:
        print(type(x))
        x = x.numpy()
    from scipy.interpolate import CubicSpline
    orig_steps = np.arange(x.shape[1])
    
    random_warps = np.random.normal(loc=1.0, scale=sigma, size=(x.shape[0], knot+2, x.shape[2]))
    warp_steps = (np.ones((x.shape[2],1))*(np.linspace(0, x.shape[1]-1., num=knot+2))).T
    
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            time_warp = CubicSpline(warp_steps[:,dim], warp_steps[:,dim] * random_warps[i,:,dim])(orig_steps)
            scale = (x.shape[1]-1)/time_warp[-1]
            ret[i,:,dim] = np.interp(orig_steps, np.clip(scale*time_warp, 0, x.shape[1]-1), pat[:,dim]).T
    return ret

def window_slice(x, reduce_ratio=0.85):
    # https:timeit magic//halshs.archives-ouvertes.fr/halshs-01357973/document
    target_len = np.ceil(reduce_ratio*x.shape[1]).astype(int)
    if target_len >= x.shape[1]:
        return x
    starts = np.random.randint(low=0, high=x.shape[1]-target_len, size=(x.shape[0])).astype(int)
    ends = (target_len + starts).astype(int)
    
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            ret[i,:,dim] = np.interp(np.linspace(0, target_len, num=x.shape[1]), np.arange(target_len), pat[starts[i]:ends[i],dim]).T
    return ret

def window_warp(x, window_ratio=0.1, scales=[0.5, 2.]):
    # https://halshs.archives-ouvertes.fr/halshs-01357973/document
    warp_scales = np.random.choice(scales, x.shape[0])

    warp_size = np.ceil(window_ratio*x.shape[1]).astype(int)
    window_steps = np.arange(warp_size)
        
    window_starts = np.random.randint(low=1, high=x.shape[1]-warp_size-1, size=(x.shape[0])).astype(int)
    window_ends = (window_starts + warp_size).astype(int)
            
    ret = np.zeros_like(x)
    for i, pat in enumerate(x):
        for dim in range(x.shape[2]):
            start_seg = pat[:window_starts[i],dim]
            window_seg = np.interp(np.linspace(0, warp_size-1, num=int(warp_size*warp_scales[i])), window_steps, pat[window_starts[i]:window_ends[i],dim])
            end_seg = pat[window_ends[i]:,dim]
            warped = np.concatenate((start_seg, window_seg, end_seg))                
            ret[i,:,dim] = np.interp(np.arange(x.shape[1]), np.linspace(0, x.shape[1]-1., num=warped.size), warped).T
    return ret

def window_warp(x, window_ratio=0.1, scales=[0.5, 2.]):
    # https://halshs.archives-ouvertes.fr/halshs-01357973/document
    warp_scales = np.random.choice(scales, x.shape[0]) #Scaling for each element in batch
     
    warp_size = np.ceil(window_ratio*x.shape[1]).astype(int) #get 0.1*window length
    window_steps = np.arange(warp_size) #number of these steps 
        
    window_starts = np.random.randint(low=1, high=x.shape[1]-warp_size-1, size=(x.shape[0])).astype(int) #Makes array of random values of where to start warp for each batch sample
    window_ends = (window_starts + warp_size).astype(int) #gets relative end points
            
    ret = np.zeros_like(x) #output array
    for i, pat in enumerate(x): #loop through samples
        for dim in range(x.shape[2]): # Loop through features
            start_seg = pat[:window_starts[i],dim]

            window_seg = np.interp(np.linspace(0, warp_size-1, num=int(warp_size*warp_scales[i])), window_steps, pat[window_starts[i]:window_ends[i],dim])

            end_seg = pat[window_ends[i]:,dim]
            warped = np.concatenate((start_seg, window_seg, end_seg))                
            ret[i,:,dim] = np.interp(np.arange(x.shape[1]), np.linspace(0, x.shape[1]-1., num=warped.size), warped).T
    return ret
timer_set = 0
def t():
  global timer_set
  print(time.time() - timer_set)
  timer_set = time.time() 

if __name__ == "__main__":
    import timeit
    
    funcs = [jitter, scaling, rotation, permutation , window_warp]
    batch_size = 512
    window_length = 500
    features = 27
    x = np.random.rand(batch_size, window_length, features)

    for func in funcs:
      print("Total time for ",func.__name__,": ", timeit.timeit("{}(x)".format(func.__name__), "from __main__ import {}, {}".format(func.__name__, "x"), number = 1) , " Seconds")
          



import json
import numpy as np 
from HPO.data.UEA_datasets import UEA_Train, UEA_Test, UEA_Full
import time
import sys
import os 
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.DARTS_utils import config_space_2_DARTS
from HPO.utils.FCN import FCN 
import pandas as pd
import torch
from HPO.data.teps_datasets import Train_TEPS , Test_TEPS
import torch.nn as nn
from torch import Tensor
from torch.utils.data import DataLoader, SubsetRandomSampler
import random
import HPO.utils.augmentation as aug
from HPO.utils.train_utils import collate_fn_padd
from HPO.utils.train import train_model, auto_train_model
from HPO.utils.weight_freezing import freeze_FCN, freeze_resnet
from HPO.utils.ResNet1d import resnet18
from HPO.utils.files import save_obj
from queue import Empty
from sklearn.model_selection import StratifiedKFold as KFold
from collections import namedtuple
from HPO.utils.worker_score import Evaluator 
from HPO.utils.worker_utils import LivePlot
from HPO.workers.worker_wrapper import __compute
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')


def compute(*args, **kwargs):
  __compute(*args, **kwargs, _compute = _compute)


def _compute(hyperparameter,cuda_device, JSON_CONFIG ):
  if "ID" in hyperparameter:
    print("ID is: {}".format(hyperparameter["ID"]))  
  ### Configuration 
  with open(JSON_CONFIG) as f:
    data = json.load(f)
    SETTINGS = data["WORKER_CONFIG"]
    SAVE_PATH = data["SEARCH_CONFIG"]["PATH"]
  
  if cuda_device == None:
     cuda_device = 1
  torch.cuda.empty_cache()

  torch.cuda.set_device(cuda_device)
    
  ##Dataset Initialisation
  #datasets = UEA_Handler("/home/cmackinnon/scripts/datasets/UEA/")
  name = SETTINGS["DATASET_CONFIG"]["NAME"]
  #train_args = [False, cuda_device ,None,1]
  # test_args = [False, cuda_device , None,1]
  if "AUGMENTATIONS" in SETTINGS:
    augs = aug.initialise_augmentations(SETTINGS["AUGMENTATIONS"])
  else: 
    augs = None 
  train_dataset = UEA_Train(name,cuda_device,augmentation = augs )
  test_dataset = UEA_Test(name,cuda_device)
  #test_dataset = datasets.load_all(name,train_args,test_args)

  
  print("Cuda Device Value: ", cuda_device)
  gen = config_space_2_DARTS(hyperparameter,reduction = True)
  print(gen)
  ##NOTE THIS NEEDS TO BE TEMP
  hyperparameter["layers"] = 1

  n_classes = train_dataset.get_n_classes()
  multibatch = False
  torch.cuda.empty_cache()
  trainloader = torch.utils.data.DataLoader(
                          train_dataset,collate_fn = collate_fn_padd,shuffle = True,
                          batch_size=SETTINGS["BATCH_SIZE"], drop_last = True)
  testloader = torch.utils.data.DataLoader(
                      test_dataset,collate_fn = collate_fn_padd,shuffle = True,
                      batch_size= SETTINGS["BATCH_SIZE"] ,drop_last = True)
  n_classes = test_dataset.get_n_classes()
  evaluator = Evaluator(SETTINGS["BATCH_SIZE"], test_dataset.get_n_classes(),cuda_device,testloader = testloader)   
  print("classes: {}".format(train_dataset.get_n_classes()))
  model = NetworkMain(train_dataset.get_n_features(),256, num_classes= train_dataset.get_n_classes(), layers = hyperparameter["layers"], auxiliary = False,drop_prob = SETTINGS["P"], genotype = gen, binary = SETTINGS["BINARY"])
  model = model.cuda(device = cuda_device)
  """
  ### Train the model
  """
  params = sum(p.numel() for p in model.parameters() if p.requires_grad)
  train_model(model , SETTINGS, trainloader , cuda_device,logger = False, evaluator = evaluator if SETTINGS["LIVE_EVAL"] else None) 
  torch.cuda.empty_cache()
  model.eval()
  evaluator.forward_pass(model, testloader,SETTINGS["BINARY"])
  evaluator.predictions(model_is_binary = SETTINGS["BINARY"] , THRESHOLD = SETTINGS["THRESHOLD"])
  total = evaluator.T()
  acc  =  evaluator.T_ACC()
  recall = evaluator.TPR(1)
  recall_total = evaluator.P(1)
  print("Accuracy: ", "%.4f" % ((acc)*100), "%")
  print("Recall: ", "%.4f" % ((recall)*100), "%")
  torch.save(model.state_dict(),"{}/weights/{:.02f}-{}".format(SAVE_PATH,acc,hyperparameter["ID"]))
  return acc, recall,params


if __name__ == "__main__":
  with open(sys.argv[1]) as f:
    HP = json.load(f)["WORKER_CONFIG"]
  _compute(HP,1,sys.argv[1])



import random
import torch
import torch.nn as nn
import numpy as np
import json
from HPO.utils.triplet import Batch_All_Triplet_Loss as Triplet_Loss
from torch import Tensor
from HPO.utils.train_log import Logger
from HPO.utils.utils import calculate_train_vals
from HPO.utils.model_constructor import Model
from torch.utils.data import DataLoader
from HPO.utils.train_utils import stdio_print_training_data
from sklearn.metrics import confusion_matrix

def stdio_print_training_data( iteration : int , outputs : Tensor, labels : Tensor , epoch : int, epochs : int, correct :int , total : int , peak_acc : float , loss : Tensor, n_iter, loss_list = None, binary = True):
  def cal_acc(y,t):
    c = np.count_nonzero(y.T==t)
    tot = len(t)
    return c , tot

  def convert_label_max_only(y):
    y = y.cpu().detach().numpy()
    idx = np.argmax(y,axis = 1)
    out = np.zeros_like(y)
    for count, i in enumerate(idx):
        out[count, i] = 1
    return idx
  if binary == True:
    new_correct , new_total = cal_acc(outputs.ge(0.5).cpu().detach().numpy(), labels.cpu().detach().numpy())
  elif len(labels.shape) > 1:
    new_correct , new_total =  cal_acc(convert_label_max_only(outputs), convert_label_max_only(labels))
  else:
    new_correct , new_total =  cal_acc(convert_label_max_only(outputs), labels.cpu().detach().numpy())
  correct += new_correct 
  total += new_total
  acc = correct / total 
  if acc > peak_acc:
    peak_acc = acc
  # Save the canvas
  print("Epoch (",str(epoch),"/",str(epochs), ") Accuracy: ","%.2f" % acc, "Iteration(s) (", str(iteration),"/",str(n_iter), ") Loss: "
    ,"%.6f" % loss," Correct / Total : {} / {} ".format(correct , total),  end = '\r')
  return correct ,total ,peak_acc

def train_model_triplet(model : Model , hyperparameter : dict, dataloader : DataLoader , epochs : int, 
    batch_size : int, cuda_device = None, augment_on = 0, graph = None, binary = False,evaluator= None,logger = None,run = None):
  if cuda_device == None:
    cuda_device = torch.cuda.current_device()
  n_iter = len(dataloader) 
  optimizer = torch.optim.Adam(model.parameters(),lr = hyperparameter["lr"])
  #scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, "max",patience = 4,verbose = True, factor = 0.1,cooldown = 2,min_lr = 0.0000000000000001)
  scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer,epochs)
  criterion = Triplet_Loss(m = 0.1 , device = cuda_device)
  epoch = 0
  peak_acc = 0
  loss_list = []
  total = 0
  correct = 0
  acc = 0
  recall = 0
  if logger == None:
    logger = Logger()
  while epoch < epochs:

    loss_values = 0
    if epoch % 3 == 0:
      total = 0
      correct = 0
    for i, (samples, labels) in enumerate( dataloader ):
      batch_size = samples.shape[0]
      optimizer.zero_grad()
      outputs = model(samples)
      if binary:
        loss = criterion(outputs.view(batch_size), labels.float()).cuda(device = cuda_device)
      else:
        loss = criterion(outputs, labels).cuda(device = cuda_device)
      loss.backward()
      optimizer.step()
      loss_values += criterion.get_fraction_pos()
      if i % 5 == 0:
        print("Epoch {} - [{}/{}] loss over epoch: {}".format(epoch,i,len(dataloader),loss_values/i))
    if epoch % 1 == 0 and run != None:
        torch.save(model.state_dict() ,"SWA/run-{}-checkpoint-{}".format(run, epoch))
    if epoch % 5 == 0:
        torch.save(model.state_dict() ,"triplet-{}".format(epoch))
        if evaluator != None:
          model.eval()
          evaluator.forward_pass(model,binary = binary)
          evaluator.predictions(model_is_binary = binary,THRESHOLD = 0.4)
          if binary:
            evaluator.ROC("train")
          acc = evaluator.T_ACC()
          recall = evaluator.TPR(1)
          evaluator.reset_cm()
          model.train()
          print("")
          print("Validation set Accuracy: {} -- Recall: {}".format(acc,recall))
          print("")
    scheduler.step()
    epoch += 1
  print()
  print("Num epochs: {}".format(epoch))
  return logger

def clone_state_dict(state_dict):
    return {name: val.clone() for name, val in state_dict.items()}

def average_state_dicts(state_dicts):
    avg_state_dict = {name: torch.stack([d[name] for d in state_dicts], dim=0).mean(dim=0)
                      for name in state_dicts[0] if state_dicts[0][name].dtype.is_floating_point}
    return avg_state_dict

def train_model(model : Model , hyperparameter : dict, dataloader : DataLoader ,
     cuda_device = None, evaluator= None,logger = None,run = None,fold = None, repeat = None):

  params = sum(p.numel() for p in model.parameters() if p.requires_grad)
  #INITIALISATION
  EPOCHS = hyperparameter["EPOCHS"]
  BATCH_SIZE = hyperparameter["BATCH_SIZE"] 
  BINARY = hyperparameter["BINARY"]
  PRINT_RATE_TRAIN = hyperparameter["PRINT_RATE_TRAIN"]
  if cuda_device == None:
    cuda_device = torch.cuda.current_device()
  n_iter = len(dataloader) 

  #CONFIGURATION OF OPTIMISER AND LOSS FUNCTION
  optimizer = torch.optim.Adam(model.parameters(),lr = hyperparameter["LR"])
  if hyperparameter["SCHEDULE"] == True:
    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer,EPOCHS,eta_min = hyperparameter["LR_MIN"])
  if hyperparameter["BINARY"] == True:
    criterion = nn.BCEWithLogitsLoss().cuda(device = cuda_device)
  else:
    criterion = nn.CrossEntropyLoss().cuda(device = cuda_device)

  #INITIALISE TRAINING VARIABLES
  epoch = 0
  peak_acc = 0
  loss_list = []
  total = 0
  correct = 0
  cm_test = ""
  acc = 0
  val_loss = torch.Tensor([0])
  recall = 0
  val_acc = 0
  if hyperparameter["LOGGING"]:
    logger = Logger(hyperparameter["database"],hyperparameter["experiment"],hyperparameter["DATASET_CONFIG"]["NAME"],fold,repeat,params)
  else:
    logger = None


  #MAIN TRAINING LOOP
  while epoch < EPOCHS:
    if epoch % 3 == 0:
      total = 0
      pred_tensor = torch.Tensor()
      gt_tensor = torch.Tensor()
      correct = 0
    weights =[]
    for i, (samples, labels) in enumerate( dataloader ):
      optimizer.zero_grad()
      #samples, labels = samples.cuda(cuda_device).float(), labels.cuda(cuda_device).long()
      outputs = model(samples)
      if BINARY == True:
        loss = criterion(outputs.view(BATCH_SIZE), labels.float())
      else:
        loss = criterion(outputs, labels)
      loss.backward()
      optimizer.step()
      if hyperparameter["WEIGHT_AVERAGING_RATE"] and epoch > EPOCHS/2:
        weights.append(clone_state_dict(model.state_dict()))
      if hyperparameter["LOGGING"]:
        pred_tensor = torch.cat((pred_tensor, outputs.detach().cpu().flatten(end_dim = 0)))
        gt_tensor = torch.cat((gt_tensor, labels.detach().cpu().flatten()))

      if PRINT_RATE_TRAIN and i % PRINT_RATE_TRAIN == 0:
        correct , total, peak_acc = stdio_print_training_data(i , outputs , labels, epoch,EPOCHS , correct , total, peak_acc, loss.item(), n_iter, loss_list,binary = BINARY)

    if hyperparameter["LOGGING"]:
      cm_train, train_acc = calculate_train_vals(pred_tensor,gt_tensor)
      if PRINT_RATE_TRAIN:
        pred_labels = torch.argmax(pred_tensor, dim=1).numpy()
        gt_labels = gt_tensor.numpy()
        with np.printoptions(linewidth = (10*len(np.unique(gt_labels))+20),precision=4, suppress=True):
          print(confusion_matrix(gt_labels,pred_labels))

    if hyperparameter["MODEL_VALIDATION_RATE"] and epoch % hyperparameter["MODEL_VALIDATION_RATE"] == 0:
        if evaluator != None:
          model.eval()
          evaluator.forward_pass(model,binary = BINARY)
          evaluator.predictions(model_is_binary = BINARY,THRESHOLD = hyperparameter["THRESHOLD"])
          val_loss = 0#evaluator.calculate_loss(criterion,BINARY)
          val_acc = evaluator.T_ACC()
          recall = evaluator.TPR(1)
          cm_test = evaluator.confusion_matrix.copy()
          bal_acc = evaluator.balanced_acc()
          evaluator.reset_cm()
          model.train()
          print("")
          print("Validation set Accuracy: {} -- Balanced Accuracy: {} -- loss: {}".format(val_acc, bal_acc ,val_loss))
          print("")

    if hyperparameter["LOGGING"]:
      cm_test = json.dumps(cm_test.tolist())
      logger.update({
        "loss": loss.item(), 
        "training_accuracy": train_acc,
        "ID":hyperparameter["ID"],
        "epoch": epoch, 
        "validation_accuracy": val_acc, 
        "lr":optimizer.param_groups[0]['lr'],
        "validation_loss": val_loss.item(),
        "confusion_matrix_train": cm_train,
        "confusion_matrix_test": cm_test
        })


    if hyperparameter["SCHEDULE"] == True:
      scheduler.step()
    epoch += 1
  if hyperparameter["WEIGHT_AVERAGING_RATE"]:  
    model.load_state_dict(average_state_dicts(weights))
  #print()
  #print("Num epochs: {}".format(epoch))
  if hyperparameter["LOGGING"]:
    logger.close()
  return logger

def train_model_regression(model : Model , hyperparameter : dict, dataloader : DataLoader ,
     cuda_device = None, evaluator= None,logger = None,run = None):

  #INITIALISATION
  EPOCHS = hyperparameter["EPOCHS"]
  VISUAL = True
  BATCH_SIZE = hyperparameter["BATCH_SIZE"] 
  BINARY = hyperparameter["BINARY"]
  PRINT_RATE_TRAIN = hyperparameter["PRINT_RATE_TRAIN"]
  if cuda_device == None:
    cuda_device = torch.cuda.current_device()
  n_iter = len(dataloader) 

  #CONFIGURATION OF OPTIMISER AND LOSS FUNCTION
  optimizer = torch.optim.Adam(model.parameters(),lr = hyperparameter["LR"])
  scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer,EPOCHS)
  criterion = nn.MSELoss().cuda(device = cuda_device)

  #INITIALISE TRAINING VARIABLES
  epoch = 0
  peak_acc = 0
  loss_list = []
  total = 0
  correct = 0
  acc = 0
  recall = 0
  if logger != False or logger == None:
    logger = Logger()
  #MAIN TRAINING LOOP
  while epoch < EPOCHS:
    if epoch % 3 == 0:
      total = 0
      correct = 0
    for i, (samples, labels) in enumerate( dataloader ):
      optimizer.zero_grad()
      #samples, labels = samples.cuda(cuda_device).float(), labels.cuda(cuda_device).long()
      outputs = model(samples)
      if BINARY == True:
        loss = criterion(outputs.view(BATCH_SIZE), labels.float())
      else:
        loss = criterion(outputs.squeeze(), labels)
      loss.backward()
      optimizer.step()

      if PRINT_RATE_TRAIN and i % PRINT_RATE_TRAIN == 0:
        correct , total, peak_acc = stdio_print_training_data(i , outputs , labels, epoch,EPOCHS , correct , total, peak_acc, loss.item(), n_iter, loss_list,binary = BINARY)
      if logger != False:
        logger.update({"loss": loss.item(), "training_accuracy": (correct/total),"index" : i,
              "epoch": epoch, "validation_accuracy": acc, "lr":optimizer.param_groups[0]['lr'],"validation recall": recall })
    if hyperparameter["WEIGHT_AVERAGING_RATE"] and epoch % hyperparameter["WEIGHT_AVERAGING_RATE"] == 0:
        torch.save(model.state_dict() ,"SWA/run-{}-checkpoint-{}".format(run, epoch))
    if hyperparameter["MODEL_VALIDATION_RATE"] and epoch % hyperparameter["MODEL_VALIDATION_RATE"] == 0 and epoch != 0:
        if evaluator != None:
          model.eval()
          evaluator.regression(model,epoch,hyperparameter)
          model.train()
    epoch += 1
    scheduler.step()
  #print()
  #print("Num epochs: {}".format(epoch))
  return logger




import matplotlib.pyplot as plt

def plot_scores(accuracy_scores):
      plt.plot(accuracy_scores)
      plt.title("Accuracy Scores of Configurations")
      plt.xlabel("Iteration")
      plt.ylabel("Accuracy")
      plt.grid()
      plt.savefig("livelogging.png",dpi = 800)
      plt.clf()






import ConfigSpace as CS 
import ConfigSpace.hyperparameters as CSH
from ConfigSpace.read_and_write import json as cs_json
import os


def nasbench301():
  current_dir = os.path.dirname(os.path.abspath(__file__))

  configspace_path = os.path.join(current_dir, 'configs/nasbench301_configspace.json')
  with open(configspace_path, "r") as f:
      json_string = f.read()
      configspace = cs_json.read(json_string)
      for i in configspace.get_conditions():
        print(i)
      input()
  return configspace 





if __name__ == "__main__":
  cs = nasbench301()
  print(cs.sample_configuration())   



from HPO.utils.visualisation import plot_scores
import csv
from ConfigSpace import ConfigurationSpace
from HPO.algorithms.algorithm_utils import train_eval
import json
def load(PATH):
    scores = []
    recall = []
    config = []
    params = []
    with open( "{}".format(PATH) , newline = "") as csvfile:
        reader = csv.reader(csvfile, delimiter = ",")
        for row in reader:
            scores.append(float(row[0]))
            recall.append(float(row[1]))
            config.append(eval("".join(row[2])))
            if len(row) == 4:
               params.append(int(row[3])) 
    error = [1-x for x in scores]
    e_min = 1
    best_list = []
    for i in error:
      if i < e_min:
        e_min = i
      best_list.append(e_min)
    return {"scores":scores,"recall":recall,"config":config,"error":error,"best":best_list ,"params": params}


def main(worker, configspace : ConfigurationSpace, json_config):
  with open(json_config) as f:
    SETTINGS = json.load(f)["SEARCH_CONFIG"]
    
  train = train_eval( worker , json_config)
  configs = load(SETTINGS["RETRAIN_PATH"])["config"]
  scores ,recall , pop= train.eval(configs)
  #print("Best Score: ", max(scores))      
  #plot_scores(scores)
  
  best_score = max(scores)
  best_config = pop[scores.index(max(scores))]
  best_rec = recall[scores.index(max(scores))]
 




import torch
import torch.nn as nn
import torch.nn.functional as F

class MultiResAtt(nn.Module):
    def __init__(self, input_channels, num_classes):
        super(MultiResAtt, self).__init__()
        num_IMUs = 1
        self.initial_block = nn.Sequential(
            nn.Conv1d(input_channels, 64, kernel_size=3, padding=1),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2)
        )

        self.residual_modules = nn.ModuleList([nn.Sequential(
            nn.Conv1d(64, 64, kernel_size=3, padding=1),
            nn.BatchNorm1d(64),
            nn.ReLU(),
            nn.Conv1d(64, 128, kernel_size=3, padding=1),
            nn.BatchNorm1d(128),
            nn.ReLU(),
            nn.Conv1d(128, 256, kernel_size=3, padding=1),
            nn.BatchNorm1d(256),
            nn.ReLU()
        ) for _ in range(num_IMUs)])

        self.adaptive_avg_pool = nn.AdaptiveAvgPool1d(output_size=1)

        self.bigru = nn.GRU(input_size=256 * num_IMUs, hidden_size=128, num_layers=2, batch_first=True, dropout=0.3, bidirectional=True)

        self.attention = nn.Sequential(
            nn.Linear(256, 128),
            nn.Tanh(),
            nn.Linear(128, 1)
        )

        self.fc = nn.Sequential(
            nn.Linear(256, 128),
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Linear(128, 128),
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Linear(128, num_classes),
        )

    def forward(self, x):
        x = self.initial_block(x)
        
        # Apply residual modules
        residual_outputs = []
        for residual_module in self.residual_modules:
            residual_outputs.append(residual_module(x))

        # Concatenate the outputs of residual modules
        x = torch.cat(residual_outputs, dim=1)
        
        x = self.adaptive_avg_pool(x).squeeze(dim=-1)

        # Add an extra dimension for the GRU input
        x = x.unsqueeze(1)
        
        gru_out, _ = self.bigru(x)
        attention_weights = self.attention(gru_out).squeeze(dim=-1)
        attention_weights = torch.softmax(attention_weights, dim=-1).unsqueeze(-1)
        attention_out = torch.bmm(gru_out.permute(0, 2, 1), attention_weights).squeeze(dim=-1)

        output = self.fc(attention_out)
        return output


import json
from HPO.utils.model_graph import ModelGraph
from HPO.searchspaces.graph_search_space import GraphConfigSpace
import numpy as np 
from HPO.data.UEA_datasets import UEA_Train, UEA_Test, UEA_Full
import time
import sys
import os 
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.DARTS_utils import config_space_2_DARTS
from HPO.utils.FCN import FCN 
import pandas as pd
import torch
from HPO.data.TEPS import TEPS 
import torch.nn as nn
from torch import Tensor
from torch.utils.data import DataLoader, SubsetRandomSampler
import random
import HPO.utils.augmentation as aug
from HPO.utils.train_utils import collate_fn_padd
from HPO.utils.train import train_model
from HPO.utils.weight_freezing import freeze_FCN, freeze_resnet
from HPO.utils.ResNet1d import resnet18
from HPO.utils.files import save_obj
from queue import Empty
from sklearn.model_selection import StratifiedKFold as KFold
from collections import namedtuple
from HPO.utils.worker_score import Evaluator 
from HPO.utils.worker_utils import LivePlot
from HPO.workers.worker_wrapper import __compute
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')


def compute(*args, **kwargs):
  __compute(*args, **kwargs, _compute = _compute)


def _compute(hyperparameter,cuda_device, JSON_CONFIG ):
  ### Configuration 
  with open(JSON_CONFIG) as f:
    data = json.load(f)
    SETTINGS = data["WORKER_CONFIG"]
    SAVE_PATH = data["SEARCH_CONFIG"]["PATH"]
  
  if cuda_device == None:
     cuda_device = 3
  torch.cuda.empty_cache()
  #print(hyperparameter)
  torch.cuda.set_device(cuda_device)
  #torch.autograd.set_detect_anomaly(True)
  #with torch.autograd.profiler.profile() as prof:
  for i in range(1):
    ##Dataset Initialisation
    #datasets = UEA_Handler("/home/cmackinnon/scripts/datasets/UEA/")
    name = SETTINGS["DATASET_CONFIG"]["NAME"]
    #train_args = [False, cuda_device ,None,1]
    # test_args = [False, cuda_device , None,1]
    if "AUGMENTATIONS" in SETTINGS:
      augs = aug.initialise_augmentations(SETTINGS["AUGMENTATIONS"])
    else: 
      augs = None 
    train_dataset = TEPS(train = True,cuda_device = cuda_device,augmentation = augs )
    test_dataset = TEPS(train = False,cuda_device = cuda_device)
    #test_dataset = datasets.load_all(name,train_args,test_args)

    
    #print("Cuda Device Value: ", cuda_device)

    n_classes = train_dataset.get_n_classes()
    multibatch = False
    torch.cuda.empty_cache()
    trainloader = torch.utils.data.DataLoader(
                            train_dataset,collate_fn = collate_fn_padd,shuffle = True,
                            batch_size=SETTINGS["BATCH_SIZE"], drop_last = True)
    testloader = torch.utils.data.DataLoader(
                        test_dataset,collate_fn = collate_fn_padd,shuffle = True,
                        batch_size= SETTINGS["BATCH_SIZE"] ,drop_last = True)
    n_classes = test_dataset.get_n_classes()
    evaluator = Evaluator(SETTINGS["BATCH_SIZE"], test_dataset.get_n_classes(),cuda_device,testloader = testloader)   
    #print("classes: {}".format(train_dataset.get_n_classes()))
    #g = GraphConfigSpace(50)
    #s = g.sample_configuration()
    #s = s[0]
    model = ModelGraph(train_dataset.get_n_features(),32,train_dataset.get_n_classes(),train_dataset.x.shape[2],hyperparameter["graph"],hyperparameter["ops"],device = cuda_device)
    if SETTINGS["COMPILE"]:
      model = torch.compile(model)
    model = model.cuda(device = cuda_device)
    """
    ### Train the model
    """
    params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print("Size: {}".format(params))
    train_model(model , SETTINGS, trainloader , cuda_device,logger = False, evaluator = evaluator if SETTINGS["LIVE_EVAL"] else None) 
  #print(prof.key_averages().table(sort_by="self_cpu_time_total"))
  torch.cuda.empty_cache()
  model.eval()
  evaluator.forward_pass(model, testloader,SETTINGS["BINARY"])
  evaluator.predictions(model_is_binary = SETTINGS["BINARY"] , THRESHOLD = SETTINGS["THRESHOLD"])
  total = evaluator.T()
  acc  =  evaluator.T_ACC()
  recall = evaluator.TPR(1)
  recall_total = evaluator.P(1)
  print("Accuracy: ", "%.4f" % ((acc)*100), "%")
  print("Recall: ", "%.4f" % ((recall)*100), "%")
  torch.save(model.state_dict(),"{}/weights/{:.02f}-{}".format(SAVE_PATH,acc,hyperparameter["ID"]))
  return acc, recall,params


if __name__ == "__main__":
  for i in range(500):
    with open(sys.argv[1]) as f:
      HP = json.load(f)["WORKER_CONFIG"]
      HP["ID"] = i
    _compute(HP,3,sys.argv[1])



import json
from HPO.utils.model_graph import ModelGraph
from HPO.searchspaces.graph_search_space import GraphConfigSpace
import numpy as np 
from HPO.data.UEA_datasets import UEA_Train, UEA_Test, UEA_Full
import time
import sys
from HPO.utils.utils import MetricLogger
import os 
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.DARTS_utils import config_space_2_DARTS
from HPO.utils.FCN import FCN 
import pandas as pd
import torch
from sklearn.model_selection import GroupKFold
from HPO.data.teps_datasets import Train_TEPS , Test_TEPS
import torch.nn as nn
from torch import Tensor
from torch.utils.data import DataLoader, SubsetRandomSampler
from torchsummary import summary
import random
import HPO.utils.augmentation as aug
from HPO.utils.train_utils import collate_fn_padd,BalancedBatchSampler, highest_power_of_two
from HPO.utils.train import train_model
from HPO.utils.weight_freezing import freeze_FCN, freeze_resnet
from HPO.utils.ResNet1d import resnet18
from HPO.utils.files import save_obj
from queue import Empty
from sklearn.model_selection import StratifiedKFold as KFold
from collections import namedtuple
from HPO.utils.worker_score import Evaluator 
from HPO.utils.worker_utils import LivePlot
from HPO.workers.worker_wrapper import __compute
from HPO.data.dataset import get_dataset
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')




def compute(*args, **kwargs):
  __compute(*args, **kwargs, _compute = _compute)
  return None


def _compute(hyperparameter,cuda_device, JSON_CONFIG):
  ### Configuration 
  if type(JSON_CONFIG) != dict:
    with open(JSON_CONFIG) as f:
      data = json.load(f)
  else:
    data = JSON_CONFIG
      
  SETTINGS = data["WORKER_CONFIG"]
  SETTINGS["ID"] = hyperparameter["ID"]
  SETTINGS["database"] = data["DATABASE_NAME"]
  SETTINGS["experiment"] = data["EXPERIMENT_NAME"]
  ARCH_SETTINGS = data["ARCHITECTURE_CONFIG"]
  SAVE_PATH = data["SEARCH_CONFIG"]["PATH"]
  acc = []
  metric_logger = MetricLogger(SAVE_PATH) 
  recall = []
  if cuda_device == None:
     cuda_device = 3
  torch.cuda.empty_cache()
  #print(hyperparameter)
  torch.cuda.set_device(cuda_device)
  #torch.autograd.set_detect_anomaly(True)
  #with torch.autograd.profiler.profile() as prof:

  ##Dataset Initialisation
  name = SETTINGS["DATASET_CONFIG"]["NAME"]
  if data["GENERATE_PARTITION"]:
    DS_PATH = SETTINGS["DATASET_CONFIG"]["DATASET_PATH"]
  else:
    DS_PATH = None
  if "AUGMENTATIONS" in SETTINGS:
    augs = aug.initialise_augmentations(SETTINGS["AUGMENTATIONS"])
  else: 
    augs = None

  train_args = {"cuda_device":cuda_device,"augmentation" : augs, "binary" :SETTINGS["BINARY"],"path" : DS_PATH}
  test_args = {"cuda_device":cuda_device,"augmentation" :None, "binary" :SETTINGS["BINARY"],"path" : DS_PATH}

  if SETTINGS["RESAMPLES"]:
    dataset = get_dataset(name,train_args,None )
    kfold = KFold(n_splits = 2, shuffle = True)
    splits = [(None,None)]*SETTINGS["RESAMPLES"]
    train_dataset = dataset
    test_dataset = dataset
  elif SETTINGS["CROSS_VALIDATION_FOLDS"] == False: 
    train_dataset, test_dataset = get_dataset(name,train_args, test_args)
    splits = [(None,None)]
    #print(train_dataset, test_dataset)
  elif SETTINGS["CROSS_VALIDATION_FOLDS"]:
    dataset, test_dataset = get_dataset(name,train_args, test_args)
    kfold = KFold(n_splits = SETTINGS["CROSS_VALIDATION_FOLDS"], shuffle = False)
    splits = kfold.split(dataset.x.cpu().numpy(),y = dataset.y.cpu().numpy())
    train_dataset = dataset
    test_dataset = dataset

  n_classes = train_dataset.get_n_classes()
  multibatch = False
  torch.cuda.empty_cache()
   
  SETTINGS["BATCH_SIZE"] = min( [highest_power_of_two(len(test_dataset)*4),  SETTINGS["BATCH_SIZE"]]  )
  for _ in range(SETTINGS["REPEAT"]):
    
    if SETTINGS["GROUPED_RESAMPLES"]:
      # Initialize GroupKFold cross-validator with desired number of splits
      kfold = GroupKFold(n_splits=5, shuffle = True)
      splits = [(None,None)]*SETTINGS["RESAMPLES"]
      train_dataset = dataset
      test_dataset = dataset

    elif SETTINGS["RESAMPLES"]:
      kfold = KFold(n_splits = 5, shuffle = True)
      splits = [(None,None)]*SETTINGS["RESAMPLES"]
      train_dataset = dataset
      test_dataset = dataset
    elif SETTINGS["CROSS_VALIDATION_FOLDS"] == False: 
      splits = [(None,None)]
    elif SETTINGS["CROSS_VALIDATION_FOLDS"]:
      kfold = KFold(n_splits = SETTINGS["CROSS_VALIDATION_FOLDS"], shuffle = True)
      splits = kfold.split(dataset.x.cpu().numpy(),y = dataset.y.cpu().numpy())
      train_dataset = dataset
      test_dataset = dataset
    #print("train",train_dataset.y.shape,train_dataset.y )
    #print("test",test_dataset.y.shape,test_dataset.y )
    for fold, (train_ids, test_ids) in enumerate(splits):    
      #print('---Fold No.--{}--------------------'.format(fold))
      torch.cuda.empty_cache()
      if SETTINGS["GROUPED_RESAMPLES"]:
         train_ids, test_ids = next(kfold.split(dataset.x.cpu().numpy(),y = dataset.y.cpu().numpy(),groups =dataset.groups))
         dataset.get_groups(train_ids,test_ids)
      elif SETTINGS["RESAMPLES"]:
        train_ids, test_ids = next(kfold.split(dataset.x.cpu().numpy(),y = dataset.y.cpu().numpy()))
      if SETTINGS["CROSS_VALIDATION_FOLDS"] or SETTINGS["RESAMPLES"]: 
        # Sample elements randomly from a given list of ids, no replacement.
        if SETTINGS["BALANCED_BATCH"]:
          train_batch_sampler = BalancedBatchSampler(dataset,SETTINGS["BATCH_SIZE"],train_ids)
          trainloader = torch.utils.data.DataLoader(dataset,collate_fn = collate_fn_padd,batch_sampler =train_batch_sampler)
        else: 
          train_subsampler = torch.utils.data.SubsetRandomSampler(train_ids)
          
          trainloader = torch.utils.data.DataLoader(
                                  dataset,collate_fn = collate_fn_padd,sampler = train_subsampler,
                                  batch_size=SETTINGS["BATCH_SIZE"], drop_last = True)
        test_subsampler = torch.utils.data.SubsetRandomSampler(test_ids)
        testloader = torch.utils.data.DataLoader(
                              dataset,collate_fn = collate_fn_padd,sampler = test_subsampler,
                              batch_size= SETTINGS["BATCH_SIZE"] ,drop_last = True)
      else:
        trainloader = torch.utils.data.DataLoader(
                                train_dataset,collate_fn = collate_fn_padd,shuffle = True,
                                batch_size=SETTINGS["BATCH_SIZE"], drop_last = True)
        testloader = torch.utils.data.DataLoader(
                            test_dataset,collate_fn = collate_fn_padd,shuffle = True,
                            batch_size= SETTINGS["BATCH_SIZE"] ,drop_last = True)
      if SETTINGS["RESAMPLES"] or SETTINGS["CROSS_VALIDATION_FOLDS"]:
        dataset.enable_augmentation(augs)
      n_classes = test_dataset.get_n_classes()
      evaluator = Evaluator(SETTINGS["BATCH_SIZE"], test_dataset.get_n_classes(),cuda_device,testloader = testloader)   
      #print("classes: {} - name: {}".format(train_dataset.get_n_classes(),name))
      #g = GraphConfigSpace(50)
      #s = g.sample_configuration()
      #s = s[0]
      if "stem" in hyperparameter["ops"]:
        stem_size = hyperparameter["ops"]["stem"]
      else:
        stem_size = ARCH_SETTINGS["STEM_SIZE"][0]
      model = ModelGraph(train_dataset.get_n_features(),stem_size,train_dataset.get_n_classes(),
          train_dataset.get_length(),hyperparameter["graph"],hyperparameter["ops"],device = cuda_device,
          binary = SETTINGS["BINARY"],dropout = SETTINGS["DROPOUT"],droppath = SETTINGS["DROPPATH"],
          raw_stem = SETTINGS["RAW_STEM"],embedding = SETTINGS["EMBEDDING"])



      files = os.listdir("{}/weights/".format(SAVE_PATH))

      valid_weights = []

      for i in files:
        splits = i.split("-")
        if splits[0] == str(hyperparameter["LOAD_ID"]):
           valid_weights.append(i)
      idx = np.argmax([ float(i.split("-")[-1]) for i in valid_weights]) #Best weights
      #idx = random.randint(0,len(valid_weights)-1)
      state_dict = torch.load("{}/weights/{}".format(SAVE_PATH,valid_weights[idx]))
      model.load_state_dict(state_dict, strict=True)

      model = model.cuda(device = cuda_device)


      params = sum(p.numel() for p in model.parameters() if p.requires_grad)
      #print("Size: {}".format(params))



      model.eval()
      if (SETTINGS["RESAMPLES"] or SETTINGS["CROSS_VALIDATION_FOLDS"] ) and SETTINGS["TEST_TIME_AUGMENTATION"] == False:
        dataset.disable_augmentation()  
      evaluator.forward_pass(model, testloader,SETTINGS["BINARY"],n_iter = SETTINGS["TEST_ITERATION"])
      evaluator.predictions(model_is_binary = SETTINGS["BINARY"] , THRESHOLD = SETTINGS["THRESHOLD"])
      total = evaluator.T()
      acc.append( evaluator.T_ACC())
      recall.append(evaluator.TPR(1))
      recall_total = evaluator.P(1)
      print("Accuracy: ", "%.4f" % ((acc[-1])*100), "%")
      #print("Recall: ", "%.4f" % ((recall[-1])*100), "%")
    acc_ = np.mean(acc)
    recall_ = np.mean(recall)
    print("Average Accuracy: ", "%.4f" % ((acc_)*100), "%")
  return acc_, recall_,params



def evaluate(config_path):
    from HPO.algorithms.algorithm_utils import load
    with open(config_path) as f:
      DATA = json.load(f)
      device = DATA["SEARCH_CONFIG"]["DEVICES"][0]
      HP = DATA["WORKER_CONFIG"]
      j = DATA
      j["WORKER_CONFIG"]["MODEL_VALIDATION_RATE"] = 10
      j["WORKER_CONFIG"]["REPEAT"] = 30
      j["WORKER_CONFIG"]["TEST_TIME_AUGMENTATION"] = False
      j["WORKER_CONFIG"]["RESAMPLES"] = False
      j["WORKER_CONFIG"]["PRINT_RATE_TRAIN"] = 50
      j["WORKER_CONFIG"]["LIVE_EVAL"] = True
      j["WORKER_CONFIG"]["BATCH_SIZE"] = 2
      if DATA["PARTITION_VAL_SET"]:
        j["WORKER_CONFIG"]["DATASET_CONFIG"]["NAME"] = "{}_Retrain".format(HP["DATASET_CONFIG"]["NAME"] )
      search = load( "{}/{}".format(DATA["SEARCH_CONFIG"]["PATH"],"evaluations.csv"))
      scores = []
      for ID in search["ID"]:
        path = "{}/{}/{}".format(DATA["SEARCH_CONFIG"]["PATH"],"metrics",ID)
        df = pd.read_csv(path)
        mu = df["accuracy"].mean()
        scores.append(mu)
      print("Best Accuracy: ",max(scores))
      #idx_max = np.argpartition(scores, -2)[-2]
      idx_max = scores.index(max(scores))
      HP["LOAD_ID"] = search["config"][idx_max]["ID"]
      HP["ID"] = "val"
      HP["graph"] = search["config"][idx_max]["graph"]
      HP["ops"] = search["config"][idx_max]["ops"]
    return _compute(HP,device,j)


if __name__ == "__main__":
    from HPO.algorithms.algorithm_utils import load
    with open(sys.argv[1]) as f:
      DATA = json.load(f)
      HP = DATA["WORKER_CONFIG"]
      j = DATA
      j["WORKER_CONFIG"]["MODEL_VALIDATION_RATE"] = 10
      j["WORKER_CONFIG"]["REPEAT"] = 30
      j["WORKER_CONFIG"]["TEST_TIME_AUGMENTATION"] = False
      j["WORKER_CONFIG"]["RESAMPLES"] = False
      j["WORKER_CONFIG"]["PRINT_RATE_TRAIN"] = 50
      j["WORKER_CONFIG"]["LIVE_EVAL"] = True
      if DATA["PARTITION_VAL_SET"]:
        j["WORKER_CONFIG"]["DATASET_CONFIG"]["NAME"] = "{}_Retrain".format(HP["DATASET_CONFIG"]["NAME"] )
      search = load( "{}/{}".format(DATA["SEARCH_CONFIG"]["PATH"],"evaluations.csv"))
      scores = []
      for ID in search["ID"]:
        path = "{}/{}/{}".format(DATA["SEARCH_CONFIG"]["PATH"],"metrics",ID)
        df = pd.read_csv(path)
        mu = df["accuracy"].mean()
        scores.append(mu)
      print("Best Accuracy: ",max(scores))
      #idx_max = np.argpartition(scores, -2)[-2]
      idx_max = scores.index(max(scores))
      HP["LOAD_ID"] = search["config"][idx_max]["ID"]
      HP["ID"] = "val"
      HP["graph"] = search["config"][idx_max]["graph"]
      HP["ops"] = search["config"][idx_max]["ops"]
    _compute(HP,3,j)



import csv
import copy
import torch
from multiprocessing import Pool
import time
import collections
import random
import matplotlib.pyplot as plt 
import numpy as np
from HPO.utils.visualisation import plot_scores
from HPO.algorithms.algorithm_utils import train_eval 
###Tounament Selection/Aging Selection
  #
  #parameters:
  # S : (int) number of solutions to sample at each iteration
  # P : population size
  # C : Total cycles of evolution 
  # parent : highest-accuracy model in sample
  # history : Record of all models
  # population : Currently active models


# IDEAS
#
# - Run aging evolution for "species" with 1 species per core allowing cross-over occationally
#g
#
#
#
#
#


class Model:
  def __init__(self, cs):
    self._arch = None
    self.accuracy = 0
    self.cs = cs

  def set_arch(self, arch):
    self._arch = arch
  def exchange_one( self , key : str ):
    new_hp = self.cs.sample_configuration().get_dictionary()[key]
    
    print("Old value: {}".format(self._arch.get_dictionary()[key]))
    self._arch[key] = new_hp
    print("New value: {}".format(self._arch.get_dictionary()[key]))
    

  def set_value(self, name, value):
    self._arch[name] = value
  def get_params(self):
    return list(self.arch().keys())
  def arch(self):
    return self._arch

def train_and_eval_population(worker, population, sample_batch, train):
  configs = []
  for model in population:
    configs.append(model.arch())
  
  acc , _rec , _config = train.eval( configs )
  for mod,result in zip(population, acc):
    mod.accuracy = result

def Mutate(cs, parent_model : Model) -> Model:
  model = copy.deepcopy(parent_model) 
  def select_operation(model):
    operation_list = []
    pre_string = "normal_cell_1_ops_"
    post_string = "_type"
    for parameter in model.arch():
      if (pre_string in parameter) and (post_string in parameter):
        operation_list.append(parameter[-6])
    
    return operation_list[random.randint(0,len(operation_list)-1)]
    
  def op_mutation(model: Model, operation_number : int) -> Model:
    parameter_name = "normal_cell_1_ops_"+operation_number+"_type"
    
    operations = cs.get_hyperparameter(parameter_name).choices
    old_op = model.arch()[parameter_name] 
   
    while model.arch()[parameter_name] == old_op:
      model.set_value(parameter_name,  operations[random.randint(0, len(operations)-1)]  )
    print("Old operation: ", old_op)
    print("New operation: ", model.arch()[parameter_name])
    return model
  
  def cont_mutation( model : Model , operation_number : int ):
    op_list = ["lr","augmentations" , "c1_weight" , "epochs" , "layers" , "channels", "T_0", "T_mult", "batch_size" ]
    model.exchange_one(random.choice(op_list))
    return model
    
    
     
  def hidden_state_mutation(parent : Model, operation_number : int) -> Model:
    #Select 1 of the inputs at random
    parameter_name = "normal_cell_1_ops_"+operation_number+"_input_"+str(random.randint(1,2))
    old_op = model.arch()[parameter_name] 
    count = 0 
    while model.arch()[parameter_name] == old_op:
      model.set_value(parameter_name,  random.randint(0,int(operation_number)-1)  )
      count+=1 
      if count > 20:
        operation_number = select_operation(model)
        if random.randint(0,5) == 5:
          return op_mutation(model, operation_number)
        parameter_name = "normal_cell_1_ops_"+operation_number+"_input_"+str(random.randint(1,2))
        old_op = model.arch()[parameter_name] 
        count = 0 
    print("Old operation: ", old_op)
    print("New operation: ", model.arch()[parameter_name])
    return model

  return random.choice([op_mutation, hidden_state_mutation , cont_mutation])(model ,select_operation(model))    


def regularized_evolution(configspace, worker , cycles, population_size, sample_size, sample_batch_size):
  """Algorithm for regularized evolution (i.e. aging evolution).

  Follows "Algorithm 1" in Real et al. "Regularized Evolution for Image
  Classifier Architecture Search".

  Args:
  cycles: the number of cycles the algorithm should run for.
  population_size: the number of individuals to keep in the population.
  sample_size: the number of individuals that should participate in each
      tournament.
  sample_batch_size: Number of children to generate before evaluating and adding to population
  Returns:
  history: a list of `Model` instances, representing all the models computed
      during the evolution experiment.
  """
  population = []
  history = []  # Not used by the algorithm, only used to report results.

  CS = configspace
  train = train_eval( worker, sample_batch_size, "RegEvo.csv"  )
  # Initialize the population with random models.
  while len(population) < population_size:
    model = Model( CS )
    model.set_arch( CS.sample_configuration() )
    population.append(model)
    history.append(model)

  train_and_eval_population(worker, population, sample_batch_size, train )

  # Carry out evolution in cycles. Each cycle produces a model and removes
  # another.
  children = []
  print_counter = 0
  while len(history) < cycles:

    print("Starting cycle", len(history) - population_size)
    # Sample randomly chosen models from the current population.

    if len(children) < sample_batch_size:
      sample = []
      while len(sample) < sample_size:
        # Inefficient, but written this way for clarity. In the case of neural
        # nets, the efficiency of this line is irrelevant because training neural
        # nets is the rate-determining step.
        candidate = random.choice(list(population))
        sample.append(candidate)

      # The parent is the best model in the sample.
      parent = max(sample, key=lambda i: i.accuracy)

      # Create the child model and store it.
      child = Mutate(CS,parent)
      children.append(child)

    else:
      train_and_eval_population(worker, children, sample_batch_size, train)
      for i in children:
        population.append(i)
        history.append(i)
        population.pop(0)
      children = []
      # Remove the oldest model.
      best = max(population, key=lambda i: i.accuracy)
      print("--Current best model--")
      print("Accuracy: ",best.accuracy)
      print("Architecture: ", best.arch())
      print("Population Size: ", len(population))
      print("Total Evaluations: ", len(history))
      print_counter +=1
      print_counter = 0
      accuracy_scores = []
      for i in history:
        accuracy_scores.append(i.accuracy)

      plot_scores(accuracy_scores)
  return history


def main(worker, configspace):
  pop_size = 100
  evaluations = 2500
  history = regularized_evolution(configspace, worker, cycles = evaluations, population_size =  pop_size, sample_size =25, sample_batch_size = 15)
  Architectures = []
  accuracy_scores = []
  generations = list(range(evaluations))
  for i in history:
    accuracy_scores.append(i.accuracy)
    Architectures.append(i.arch)
  
  plt.scatter(generations[:pop_size],accuracy_scores[:pop_size], c = "red")
  plt.scatter(generations[pop_size:],accuracy_scores[pop_size:])
  plt.title("Accuracy Scores of Configurations")
  plt.xlabel("Generation")
  plt.ylabel("Accuracy")
  plt.grid()
  plt.savefig("regevo.png",dpi = 1200)

  indexs = accuracy_scores.index(max(accuracy_scores))
  print("Best accuracy: ", accuracy_scores[indexs])
  print("Best Hyperparameters: ", Architectures[indexs]())




if __name__ == '__main__':
  main()



import torch
import numpy as np
import math
import torch.nn as nn
import torch.nn.functional as F
OPS = {
  'none' : lambda C, stride, affine: Zero(stride),
  'avg_pool_3x3' : lambda C, stride, affine: nn.AvgPool1d(3, stride=stride, padding=1, count_include_pad=False),
  'max_pool_3x3' : lambda C, stride, affine: nn.MaxPool1d(3, stride=stride, padding=1),
  'avg_pool_31x31' : lambda C, stride, affine: nn.AvgPool1d(31, stride=stride, padding=15, count_include_pad=False),
  'max_pool_31x31' : lambda C, stride, affine: nn.MaxPool1d(31, stride=stride, padding=15),
  'avg_pool_64x64' : lambda C, stride, affine: nn.AvgPool1d(63, stride=stride, padding=31, count_include_pad=False),
  'max_pool_64x64' : lambda C, stride, affine: nn.MaxPool1d(63, stride=stride, padding=31),
  'skip_connect' : lambda C, stride, affine: Identity() if stride == 1 else FactorizedReduce(C, C, affine=affine),
  'memory_cell'  : lambda C, stride, affine: MemoryCell(),
  'point_conv' : lambda C, stride, affine: PointConv(C, stride,affine = affine),
  'depth_conv_7' :  lambda C, stride, affine: DepthConv(C,7, stride = stride ,padding = 3, affine = affine),
  'depth_conv_15' :  lambda C, stride, affine: DepthConv(C,15, stride = stride ,padding = 7, affine = affine),
  'depth_conv_29' :  lambda C, stride, affine: DepthConv(C,29, stride = stride ,padding = 14, affine = affine),
  'depth_conv_61' :  lambda C, stride, affine: DepthConv(C,61, stride = stride ,padding = 30, affine = affine),
  'depth_conv_101' :  lambda C, stride, affine: DepthConv(C,101, stride = stride ,padding = 50, affine = affine),
  'depth_conv_201' :  lambda C, stride, affine: DepthConv(C,201, stride = stride ,padding = 100, affine = affine),
  'patch_29' : lambda C, stride, affine: Patches(C,29,stride,padding = 14,affine = affine),
  'patch_61' : lambda C, stride, affine: Patches(C,61,stride,padding = 30,affine = affine),
  'SE_8' : lambda C, stride, affine: SE(C,8,stride = stride,affine = affine),
  'SE_16' : lambda C, stride, affine: SE(C,16,stride = stride,affine = affine),
  'SE_32' : lambda C, stride, affine: SE(C,32,stride = stride,affine = affine),
  'channel_mlp' : lambda C, stride, affine: MLPChannel(C,2,stride = stride,affine = affine),
  'attention_channel' : lambda C, stride, affine: AttentionChannel(C,stride = stride),
  'attention_space' : lambda C, stride, affine: AttentionSpace(C,stride = stride),
  
  'sep_conv_3x3' : lambda C, stride, affine: SepConv(C, C, 3, stride, 1, affine=affine),
  'sep_conv_5x5' : lambda C, stride, affine: SepConv(C, C, 5, stride, 2, affine=affine),
  'sep_conv_7x7' : lambda C, stride, affine: SepConv(C, C, 7, stride, 3, affine=affine),
  'sep_conv_15x15' : lambda C, stride, affine: SepConv(C, C, 15, stride, 7, affine=affine),
  'sep_conv_30x30' : lambda C, stride, affine: SepConv(C, C, 27, stride, 13, affine=affine),
  'dil_conv_3x3' : lambda C, stride, affine: DilConv(C, C, 3, stride, 2, 2, affine=affine),
  'dil_conv_5x5' : lambda C, stride, affine: DilConv(C, C, 5, stride, 4, 2, affine=affine),
  'dil_conv_7x7' : lambda C, stride, affine: DilConv(C, C, 3, stride, 2, 2, affine=affine),
}
class ReLUConvBN(nn.Module):

  def __init__(self, C_in, C_out, kernel_size, stride, padding, affine=True):
    super(ReLUConvBN, self).__init__()
    self.op = nn.Sequential(
      nn.ReLU(inplace=False),
      nn.Conv1d(C_in, C_out, kernel_size, stride=stride, padding=padding, bias=False),
      nn.BatchNorm1d(C_out, affine=affine)
    )

  def forward(self, x):
    return self.op(x)

class Patches(nn.Module):
  def __init__(self, c,patch_size ,stride, padding, affine=True):
    super(Patches, self).__init__()
    self.conv = nn.Conv1d(c,c*patch_size, kernel_size=patch_size, stride=patch_size*stride, padding=padding, bias=False)
    self.act = nn.GELU()
    self.norm = nn.BatchNorm1d(c, affine=affine)

  def forward(self, x):
    x =self.conv(x)
    x = self.act(x)
    x = self.norm(x)
    return x

class MixMLP(nn.Module):
  def __init__(self, C, stride,patch_size,affine = False):
    super(MixMLP,self).__init__()
 
    #self.norm = nn.LayerNorm()
    embedded_size = C*patch_size
    self.patch_size = patch_size
    self.mlp1 = nn.Linear(C*patch_size, embedded_size)
    self.ge = nn.GELU()
    self.mlp2 = nn.Linear(embedded_size, (C*patch_size)/stride)
  def forward(self, x):
    #Patches
    num_patches = x.shape[2]/self.patch_size
    patches = torch.chunk(x, num_patches,dim = 2)
    
    self.mlp1 

#self.patch = Patches(c,p,stride,padding)
class Attention(nn.Module):
  def __init__(self,max_len = 960,embedding = 200):
    super(Attention,self).__init__()
    self.sqrt_dk = np.sqrt(max_len)
    self.max_len = max_len
    self.ge = nn.GELU()
    self.q_w = nn.Linear(max_len, embedding,bias = False)
    self.v_w = nn.Linear(max_len, embedding,bias = False)
    self.k_w = nn.Linear(max_len, embedding,bias = False)
  def forward(self,x):
    pad_length = self.max_len - x.shape[2]
    x = F.pad(input = x, pad = (0,pad_length),value = 0)
    query = self.ge(self.q_w(x))
    key = self.ge(self.k_w(x))
    value = self.ge(self.v_w(x))
    score = torch.bmm(query,key.transpose(1,2))/self.sqrt_dk
    attn = nn.functional.softmax(score, dim = -1)
    out = torch.bmm(attn,value)
    return out
    
class AttentionSpace(nn.Module):
  def __init__(self,C,stride):
    super(AttentionSpace,self).__init__()
    d_k = C

    self.q_w = nn.Conv1d(C, d_k,bias = False,groups = C,kernel_size = 7,padding = 3,stride = stride)
    self.v_w = nn.Conv1d(C, d_k,bias = False,groups = C,kernel_size = 7,padding = 3,stride = stride)
    self.k_w = nn.Conv1d(C, d_k,bias = False,groups = C,kernel_size = 7,padding = 3,stride = stride)
    self.sqrt_dk = np.sqrt(d_k)
  def forward(self,x):
    #print("Incoming shape of x: {}".format(x.shape))
    query = self.q_w(x)   
    key = self.k_w(x)   
    value = self.v_w(x)
    #print("Query: {}".format(query.shape))
    score = torch.bmm(query,key.transpose(1,2))/self.sqrt_dk
    attn = nn.functional.softmax(score, dim = -1)
    out = torch.bmm(attn,value)
    return out

class AttentionChannel(nn.Module):
  def __init__(self,C,stride):
    super(AttentionChannel,self).__init__()
    d_k = C

    self.q_w = nn.Conv1d(C, d_k,bias = False,kernel_size = 1,padding = 0,stride = stride)
    self.v_w = nn.Conv1d(C, d_k,bias = False,kernel_size = 1,padding = 0,stride = stride)
    self.k_w = nn.Conv1d(C, d_k,bias = False,kernel_size = 1,padding = 0,stride = stride)
    self.sqrt_dk = np.sqrt(d_k)
  def forward(self,x):
    #print("Incoming shape of x: {}".format(x.shape))
    query = self.q_w(x)   
    key = self.k_w(x)   
    value = self.v_w(x)
    #print("Query: {}".format(query.shape))
    score = torch.bmm(query,key.transpose(1,2))/self.sqrt_dk
    attn = nn.functional.softmax(score, dim = -1)
    out = torch.bmm(attn,value)
    return out
    
class SE(nn.Module):
  def __init__(self, C,r,stride,affine = True ):
    super(SE,self).__init__()
    self.GP = nn.AdaptiveAvgPool1d(1)
    self.fc1 = nn.Linear(C, C//r, bias = False)
    self.act = nn.GELU()
    self.fc2 = nn.Linear(C//r, C ,bias = False)
    self.sig = nn.Sigmoid()
    self.stride = stride
  def forward(self,x):
    #Squeeze
    y = self.GP(x).squeeze()# [Batch,C]
    #torch.mean(x,axis = 2)  
    
    y = self.fc1(y)
    y = self.act(y)
    y = self.fc2(y)
    y = self.sig(y).unsqueeze(dim = 2)
    x = x[:,:,::self.stride] 
    return x* y.expand_as(x)

class MLPChannel(nn.Module):
  def __init__(self, C,r,stride,affine = True ):
    super(MLPChannel,self).__init__()
    self.GP = nn.AdaptiveAvgPool1d(1)
    self.fc1 = nn.Linear(C, C*r, bias = False)
    self.act = nn.GELU()
    self.fc2 = nn.Linear(C*r, C ,bias = False)
    self.sig = nn.Sigmoid()
    self.stride = stride
    if self.stride > 1:
      self.reduce = nn.AvgPool1d(kernel_size = self.stride,stride = self.stride,padding = 0)
      
  def forward(self,x):
    if self.stride > 1:
      x = self.reduce(x)
    #Squeeze
    y = self.GP(x).squeeze()# [Batch,C]
    #torch.mean(x,axis = 2)  
    
    y = self.fc1(y)
    y = self.act(y)
    y = self.fc2(y)
    y = self.sig(y).unsqueeze(dim = 2)
    return x* y.expand_as(x)

class SE_mask(nn.Module):
  def __init__(self, C,r,stride,affine = True ):
    super(SE,self).__init__()
    self.GP = nn.AdaptiveAvgPool1d(1)
    self.fc1 = nn.Linear(C, C//r, bias = False)
    self.act = nn.GELU()
    self.fc2 = nn.Linear(C//r, C ,bias = False)
    self.sig = nn.Sigmoid()
    self.stride = stride
  def forward(self,x):
    #Squeeze
    y = self.GP(x).squeeze()# [Batch,C]
    #torch.mean(x,axis = 2)  
    
    y = self.fc1(y)
    y = self.act(y)
    y = self.fc2(y)
    y = self.sig(y).unsqueeze(dim = 2)
    x = x[:,:,::self.stride] 
    return x * y.expand_as(x)


class DepthConv(nn.Module):
  def __init__(self, c, kernel_size, padding,stride, affine=True):
    super(DepthConv, self).__init__()
    self.conv = nn.Conv1d(c,c, kernel_size=kernel_size, stride=stride, padding=padding, groups=c, bias=False)
    self.act = nn.GELU()
    self.norm = nn.BatchNorm1d(c, affine=affine)

  def forward(self, x):
    x =self.conv(x)
    x = self.act(x)
    x = self.norm(x)
    return x

class PointConv(nn.Module):
  def __init__(self, C, stride,affine=True):
    super(PointConv, self).__init__()
    self.conv = nn.Conv1d(C,C, kernel_size=1, stride=stride, bias=False)
    self.act = nn.GELU()
    self.norm = nn.BatchNorm1d(C, affine=affine)

  def forward(self, x):
    x =self.conv(x)
    x = self.act(x)
    x = self.norm(x)
    return x

class MemoryCell(nn.Module):
  def __init__(self):
    super(MemoryCell,self).__init__()
    self.cell = None

  def forward(self,x):
    if self.cell == None:
      self.cell = torch.ones(size = x.shape)
    out = self.cell.clone()
    self.cell = x
    return out

class DilConv(nn.Module):
    
  def __init__(self, C_in, C_out, kernel_size, stride, padding, dilation, affine=True):
    super(DilConv, self).__init__()
    self.op = nn.Sequential(
      nn.ReLU(inplace=False),
      nn.Conv1d(C_in, C_in, kernel_size=kernel_size, stride=stride, padding=padding, dilation=dilation, groups=C_in, bias=False),
      nn.Conv1d(C_in, C_out, kernel_size=1, padding=0, bias=False),
      nn.BatchNorm1d(C_out, affine=affine),
      )

  def forward(self, x):
    return self.op(x)


class SepConv(nn.Module):
    
  def __init__(self, C_in, C_out, kernel_size, stride, padding, affine=True):
    super(SepConv, self).__init__()
    self.op = nn.Sequential(
      nn.ReLU(inplace=False),
      nn.Conv1d(C_in, C_in, kernel_size=kernel_size, stride=stride, padding=padding, groups=C_in, bias=False),
      nn.Conv1d(C_in, C_in, kernel_size=1, padding=0, bias=False),
      nn.BatchNorm1d(C_in, affine=affine),
      nn.ReLU(inplace=False),
      nn.Conv1d(C_in, C_in, kernel_size=kernel_size, stride=1, padding=padding, groups=C_in, bias=False),
      nn.Conv1d(C_in, C_out, kernel_size=1, padding=0, bias=False),
      nn.BatchNorm1d(C_out, affine=affine),
      )

  def forward(self, x):
    return self.op(x)


class Identity(nn.Module):

  def __init__(self):
    super(Identity, self).__init__()

  def forward(self, x):
    return x


class Zero(nn.Module):

  def __init__(self, stride):
    super(Zero, self).__init__()
    self.stride = stride

  def forward(self, x):
    if self.stride == 1:
      return x.mul(0.)
    return x[:,:,::self.stride].mul(0.)

"""
class FactorizedReduce(nn.Module):

  def __init__(self, C_in, C_out, affine=True):
    super(FactorizedReduce, self).__init__()
    assert C_out % 2 == 0
    self.relu = nn.ReLU(inplace=False)
    self.conv_1 = nn.Conv1d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False)
    self.conv_2 = nn.Conv1d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False) 
    self.bn = nn.BatchNorm1d(C_out, affine=affine)

  def forward(self, x):
    x = self.relu(x)
    out = torch.cat([self.conv_1(x), self.conv_2(x[:,:,1:])], dim=1)
    out = self.bn(out)
    return out
"""
class FactorizedReduce(nn.Module):
    def __init__(self, C_in, C_out, affine=False):
        super().__init__()
        self.conv1 = nn.Conv1d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False)
        self.conv2 = nn.Conv1d(C_in, math.ceil(C_out / 2), 1, stride=2, padding=0, bias=False)
        self.bn = nn.BatchNorm1d(C_out, affine=affine)

    def forward(self, x):
        out = torch.cat([self.conv1(x), self.conv2(x[:,:,:])], dim=1)
        out = self.bn(out)
        return out



import torch
from HPO.utils.model_graph import ModelGraph
import torch.nn.functional as F
from HPO.data.UEA_datasets import UEA_Train, UEA_Test, UEA_Full
import time
import matplotlib.pyplot as plt
from HPO.utils.FCN import FCN 
import pandas as pd
import torch
from HPO.data.teps_datasets import Train_TEPS , Test_TEPS
import torch.nn as nn
from torch import Tensor
from torch.utils.data import DataLoader, SubsetRandomSampler
import random
import HPO.utils.augmentation as aug
from HPO.utils.train_utils import collate_fn_padd
from HPO.utils.train import train_model
from queue import Empty
from sklearn.model_selection import StratifiedKFold as KFold
from collections import namedtuple
import numpy as np 
import torch.nn as nn
import os
import sys
import copy
from HPO.utils.model import NetworkMain
from sklearn.metrics import confusion_matrix
from HPO.utils.DARTS_utils import config_space_2_DARTS
import json
from HPO.data.UEA_datasets import UEA_Train, UEA_Test, UEA_Full
import csv
from HPO.data.dataset import get_dataset
import HPO.utils.augmentation as aug

class EnsembleManager:
        def __init__(self , JSON_CONFIG,device):
                with open(JSON_CONFIG,"r") as f:
                        data = json.load(f)
                self.path = JSON_CONFIG[:-18]
                self.cuda_device =device
                self.accuracy,  self.recall, self.configs = self.load_hps(self.path)
                self.SETTINGS = data["WORKER_CONFIG"]
                self.channels = data["ARCHITECTURE_CONFIG"]["STEM_SIZE"][0]
                name = self.SETTINGS["DATASET_CONFIG"]["NAME"]
                if data["GENERATE_PARTITION"]:
                        DS_PATH = self.SETTINGS["DATASET_CONFIG"]["DATASET_PATH"]
                else:
                        DS_PATH = None

                if "AUGMENTATIONS" in self.SETTINGS:
                        augs = aug.initialise_augmentations({'Crop_1': {'rate': 0.5, 'crop_min': 0.8}})
                else: 
                        augs = None

                train_args = {"cuda_device":device,"augmentation" : augs, "binary" :self.SETTINGS["BINARY"],"path" : DS_PATH}
                test_args = {"cuda_device":device,"augmentation" :None, "binary" :self.SETTINGS["BINARY"],"path" : DS_PATH}
                train_dataset, self.test_dataset = get_dataset(name,train_args, test_args)
                train_dataset, self.aug_test_dataset = get_dataset(name,train_args, train_args)
                # = UEA_Test(name = self.SETTINGS["DATASET_CONFIG"]["NAME"],device = 0)
                self.num_classes = self.test_dataset.get_n_classes()
                self.num_features = self.test_dataset.get_n_features()
                self.models = nn.ModuleList()

        def evaluate_old(self, batch_size):
                self.testloader = torch.utils.data.DataLoader(
                      self.test_dataset,collate_fn = collate_fn_padd,shuffle = True,
                      batch_size= batch_size ,drop_last = True)
                self.labels = np.zeros(shape = len(self.test_dataset))
                self.preds = np.zeros(shape = len(self.test_dataset))
                self.ensemble.eval()
                for index, (x,y) in enumerate(self.testloader):
                        self.labels[ index*batch_size :(index+1)*batch_size] = y.detach().cpu().numpy()
                        self.preds[ index*batch_size :(index+1)*batch_size] = self.ensemble(x.float(),y).detach().cpu().numpy()
                self.confusion_matrix = confusion_matrix(self.labels,self.preds,labels = list(range(self.num_classes)))
                with np.printoptions(linewidth = (10*self.num_classes+20),precision=4, suppress=True):
                    print(self.confusion_matrix)
                correct = np.sum(np.diag(self.confusion_matrix))
                total = np.sum(self.confusion_matrix)
                print("Accuracy: {}".format(correct/total))


        def evaluate_aug(self, batch_size, n_iter=10):
                    self.testloader = torch.utils.data.DataLoader(
                        self.test_dataset, collate_fn=collate_fn_padd, shuffle=True, 
                        batch_size=batch_size, drop_last=True)

                    self.aug_testloader = torch.utils.data.DataLoader(
                        self.aug_test_dataset, collate_fn=collate_fn_padd, shuffle=False, 
                        batch_size=batch_size, drop_last=True)

                    labels_list = []
                    preds_list = []
                    self.ensemble.eval()
                    probs = torch.zeros((n_iter,len(self.aug_testloader)*batch_size,self.num_classes,len(self.ensemble.classifiers)))
                    for i in range(n_iter):
                        labels_list = []
                        for index, (x, y) in enumerate(self.aug_testloader):
                            labels_list.append(y.detach())  # append the tensor to the list directly
                            x, y = x.float(), y  # moving to the device
                            #if i == n_iter-1:

                            probs[i,batch_size*index:batch_size*(index+1),:,:] = self.ensemble.forward_aug(x, y).detach()  # append the tensor to the list directly

                            # Average predictions over the augmented samples and append to preds_list
                            #preds_list.append(torch.mean(torch.stack(aug_preds_list), dim=0))
                    print(probs.shape)
                    preds = self.ensemble.soft_voting(torch.mean(probs,axis = 0))
                    # concatenate all tensors along the 0-th dimension
                    labels = torch.cat(labels_list).cpu().numpy()
                    #preds = torch.cat(preds_list).cpu().numpy()

                    self.confusion_matrix = confusion_matrix(labels, preds, labels=list(range(self.num_classes)))
                    with np.printoptions(linewidth=(20 * self.num_classes + 20), precision=4, suppress=True):
                        print(self.confusion_matrix)
                    correct = np.sum(np.diag(self.confusion_matrix))
                    total = np.sum(self.confusion_matrix)
                    print("Accuracy: {}".format(correct / total))
                    return correct / total

        def evaluate(self, batch_size):
            self.testloader = torch.utils.data.DataLoader(
                self.test_dataset, collate_fn=collate_fn_padd, shuffle=True, 
                batch_size=batch_size, drop_last=False)

            labels_list = []
            preds_list = []
            self.ensemble.eval()
            
            for index, (x, y) in enumerate(self.testloader):
                x, y = x.float(), y  # moving to the device
                labels_list.append(y.detach())  # append the tensor to the list directly
                preds_list.append(self.ensemble(x, y).detach())  # append the tensor to the list directly

            # concatenate all tensors along the 0-th dimension
            labels = torch.cat(labels_list).cpu().numpy()
            preds = torch.cat(preds_list).cpu().numpy()
            print(labels.shape)

            self.confusion_matrix = confusion_matrix(labels, preds, labels=list(range(self.num_classes)))
            with np.printoptions(linewidth=(10 * self.num_classes + 20), precision=4, suppress=True):
                print(self.confusion_matrix)
            correct = np.sum(np.diag(self.confusion_matrix))
            total = np.sum(self.confusion_matrix)
            print("Accuracy: {}".format(correct / total))
            return correct / total


        def get_ensemble(self,n_classifiers = 10):
                order = list(np.argsort(self.accuracy))
                """
                weights = os.listdir("{}weights/".format(self.path))
                accs = [ float(i.split("-")[-1]) for i in weights]
                order =  list(np.argsort(accs))
                """

                while len(self.models) < n_classifiers:
                        index = order.pop(-1)
                        print("Model acc: {}".format(self.accuracy[index]))
                        m, sucess = self.try_build_max(index)
                        if sucess:
                          self.models.extend(m)
                self.ensemble = Ensemble(self.models,self.num_classes,self.cuda_device)


        def load_hps(self, PATH,FILENAME = "evaluations.csv"):
            scores = []
            recall = []
            config = []
            with open( "{}{}".format(PATH,FILENAME) , newline = "") as csvfile:
                    reader = csv.reader(csvfile, delimiter = ",")
                    for row in reader:
                        scores.append(float(row[0]))
                        recall.append(float(row[1]))
                        config.append(eval("".join(row[2])))
            return scores, recall, config
    
        def find_all(self,acc):
            current_val = []
            for idx,i in enumerate(self.accuracy):
                if acc == i:
                        current_val.append(idx)
            return current_val

        def try_build(self,index):
          try:
            hyperparameter = self.configs[index]
            print(hyperparameter)
            ID = hyperparameter["ID"]
            weights = os.listdir("{}weights/".format(self.path))
            num_len = len(str(ID))
            models = []
            for inst in weights:
              if int(inst.split("-")[0]) == int(ID):
                print(inst,ID)
                match = inst

                state = torch.load("{}weights/{}".format(self.path,match))
                if "graph" in hyperparameter.keys():
                        print("loading graph")
                        models.append(ModelGraph(self.test_dataset.get_n_features(),self.channels,self.test_dataset.get_n_classes(),
                          self.test_dataset.get_length(),hyperparameter["graph"],hyperparameter["ops"],device = self.cuda_device,
                          binary = self.SETTINGS["BINARY"],dropout = self.SETTINGS["DROPOUT"],droppath = self.SETTINGS["DROPPATH"],
                          raw_stem = self.SETTINGS["RAW_STEM"],embedding = self.SETTINGS["EMBEDDING"]))
                        models[-1].load_state_dict(state)
                else: 
                        print("loading cell")
                        gen = config_space_2_DARTS(hyperparameter, reduction = True)
                        #print("Loaded Weights")
                        #print("Weight mismatch")
                        models.append(NetworkMain(self.num_features,
                                    2**hyperparameter["channels"], num_classes= self.num_classes,
                                  layers = hyperparameter["layers"], auxiliary = False,
                                  drop_prob = self.SETTINGS["P"],genotype = gen, 
                                  binary = self.SETTINGS["BINARY"]).load_state_dict(state))
            return models,True
          except:
            return None, False

        def try_build_max(self,index):
            try:
                hyperparameter = self.configs[index]
                print(hyperparameter)
                ID = hyperparameter["ID"]
                weights = os.listdir("{}weights/".format(self.path))
                num_len = len(str(ID))
                models = []
                max_acc_for_fold = {}  # Dictionary to store max accuracy for each fold
                best_weights_for_fold = {}  # Dictionary to store weight file with max accuracy for each fold
                for inst in weights:
                    if int(inst.split("-")[0]) == int(ID):
                        print(inst,ID)
                        fold = int(inst.split("-")[1])  # Assuming fold is second part of the filename
                        acc = float(inst.split("-")[2])  # Assuming accuracy is third part of the filename
                        # If this fold isn't in the dictionary yet or this acc is higher than the current max
                        if fold not in max_acc_for_fold or acc > max_acc_for_fold[fold]:
                            max_acc_for_fold[fold] = acc  # Update max acc for this fold
                            best_weights_for_fold[fold] = inst  # Update weight file for this fold

                for fold, weight_file in best_weights_for_fold.items():
                    state = torch.load("{}weights/{}".format(self.path, weight_file))
                    if "graph" in hyperparameter.keys():
                            print("loading graph")
                            print("Accuracy: {}".format(weight_file.split("-")[-1]))
                            models.append(ModelGraph(self.test_dataset.get_n_features(),self.channels,self.test_dataset.get_n_classes(),
                              self.test_dataset.get_length(),hyperparameter["graph"],hyperparameter["ops"],device = self.cuda_device,
                              binary = self.SETTINGS["BINARY"],dropout = self.SETTINGS["DROPOUT"],droppath = self.SETTINGS["DROPPATH"],
                              raw_stem = self.SETTINGS["RAW_STEM"],embedding = self.SETTINGS["EMBEDDING"]))
                            models[-1].load_state_dict(state)
                    else: 
                            print("loading cell")
                            gen = config_space_2_DARTS(hyperparameter, reduction = True)
                            models.append(NetworkMain(self.num_features,
                                        2**hyperparameter["channels"], num_classes= self.num_classes,
                                      layers = hyperparameter["layers"], auxiliary = False,
                                      drop_prob = self.SETTINGS["P"],genotype = gen, 
                                      binary = self.SETTINGS["BINARY"]).load_state_dict(state))
                return models,True
            except:
                return None, False


        def load_state(path,ID):
                state = torch.load("{}{}".format(path,ID))
        


class Ensemble(nn.Module):
        def __init__( self, models,num_classes ,device):
                super(Ensemble,self).__init__()
                
                self.num_classes = num_classes
                self.classifiers = models.cuda(device)

        def soft_voting( self, prob ):
                #SUM PROBABILITIES FROM ALL CLASSIFIERS THEN GET MAX PROBABILITY DENSITY
                t_prob = torch.sum(prob, axis =  2)
                preds = torch.argmax(t_prob,dim = 1)
                return preds #shape [batch_size] 
        def hard_voting(self, prob):
                #GET PREDICTION FROM EACH MODEL THEN CALCULATE MODE (MAJORITY)
                preds = torch.argmax(prob , axis = 1)
                return torch.mode(preds,dim = 1)[0] #Shape [batch_size]

        def forward_aug( self, x ,y):
                self.batch_size = x.shape[0]
                probs = torch.zeros(size = (self.batch_size ,self.num_classes, len(self.classifiers) ))
                for idx , model in enumerate(self.classifiers):
                        probs[:, :, idx] = F.sigmoid(model(x))
                #print(probs)
                #print(F.sigmoid(probs))
                #print(y)
                return probs

        def forward( self, x ,y):
                self.batch_size = x.shape[0]
                probs = torch.zeros(size = (self.batch_size ,self.num_classes, len(self.classifiers) ))
                for idx , model in enumerate(self.classifiers):
                        probs[:, :, idx] = F.sigmoid(model(x))
                #print(probs)
                #print(F.sigmoid(probs))
                #print(y)
                return self.soft_voting(probs)

        def eval(self):
                for i in self.classifiers:
                        i.eval()

if __name__ == "__main__":
    import sys
    be = EnsembleManager(sys.argv[1],1)
    be.get_ensemble(50)
    be.evaluate(4)



from HPO.utils.visualisation import plot_scores
import csv
from ConfigSpace import ConfigurationSpace
from HPO.algorithms.algorithm_utils import train_eval,load
import json
import copy
import random
import time

def load_scores(config):
  path = "{}/{}".format(SETTINGS["PATH"],"metrics/")


def main(worker, configspace : ConfigurationSpace, json_config):
  with open(json_config) as f:
    SETTINGS = json.load(f)["SEARCH_CONFIG"]
  
  TIME_SINCE_IMPROVE = 0
  EARLY_STOP = SETTINGS["EARLY_STOP"]
  train = train_eval( worker , json_config)
  if SETTINGS["RESUME"]:
    data = load(SETTINGS["EXPERIMENT_NAME"])
    history_scores = data["scores"]
    history_conf = data["config"]
  else:
    configs = configspace.sample_configuration(SETTINGS["INITIAL_POPULATION_SIZE"])
    scores , recall , pop= train.init_async(configs)
    history_scores = scores
    history_conf = pop
    last_max_indexs = None
    iteration = 0
  while iteration < SETTINGS["TOTAL_EVALUATIONS"]:
    scores ,recall , pop = [], [], []
    while len(scores) == 0:
      time.sleep(0.5)
      scores ,recall , pop = train.get_async()
    history_conf.extend(pop)
    history_scores.extend(scores)
    print("Generation: {}".format(iteration))
    best_score = max(history_scores)
    max_indices = [index for index, value in enumerate(history_scores) if value == best_score]
    #idx_best = history_scores.index(best_score)
    print("Best Score: {}".format(best_score))
    best_configs = [history_conf[i] for i in max_indices ]
    configs = []
    while train.config_queue.qsize() < SETTINGS["CORES"]/2:
      train.update_async(configspace.mutate_graph(random.choice(best_configs),2))
    if max_indices == last_max_indexs:
      TIME_SINCE_IMPROVE+=len(scores)
    else:
      TIME_SINCE_IMPROVE = 0
    last_max_indexs = copy.copy(max_indices)
    if TIME_SINCE_IMPROVE > EARLY_STOP:
      break
    iteration+= 1

  #plot_scores(scores)
  
  best_score = max(scores)
  best_config = pop[scores.index(max(scores))]
  best_rec = recall[scores.index(max(scores))]

def main_batch(worker, configspace : ConfigurationSpace, json_config):
  with open(json_config) as f:
    SETTINGS = json.load(f)["SEARCH_CONFIG"]
  
  GENERATIONS = int(SETTINGS["TOTAL_EVALUATIONS"] / SETTINGS["CORES"])
  TIME_SINCE_IMPROVE = 0
  EARLY_STOP = 150
  train = train_eval( worker , json_config)
  configs = configspace.sample_configuration(SETTINGS["CORES"])
  scores ,recall , pop= train.eval(configs)
  history_scores = scores
  history_conf = pop
  last_max_indexs = None
  iteration = 0
  while True:
    print("Generation: {}".format(iteration))
    best_score = max(history_scores)
    max_indices = [index for index, value in enumerate(history_scores) if value == best_score]
    #idx_best = history_scores.index(best_score)
    print("Best Score: {}".format(best_score))
    best_configs = [history_conf[i] for i in max_indices ]
    configs = []
    for i in range(SETTINGS["CORES"]):
      configs.append(configspace.mutate_graph(random.choice(best_configs)))
      print("Got {} configuration".format(len(configs)))
    scores ,recall , pop= train.eval(configs)
    history_conf.extend(pop)
    history_scores.extend(scores)
    if max_indices == last_max_indexs:
      TIME_SINCE_IMPROVE+=SETTINGS["CORES"]
    else:
      TIME_SINCE_IMPROVE = 0
    last_max_indexs = copy.copy(max_indices)
    if TIME_SINCE_IMPROVE > EARLY_STOP:
      break
    iteration+= 1

  #plot_scores(scores)
  
  best_score = max(scores)
  best_config = pop[scores.index(max(scores))]
  best_rec = recall[scores.index(max(scores))]
 



import numpy as np
import os
import torch
from torch.utils.data import Dataset
import random 
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
"""
class BTC(Dataset):
  def __init__(self, window_size, prediction_distance, split = 5,augmentation = False,samples_per_class = None ,device = None):
    self.path ="{}/scripts/datasets/BTC/BTC.npy".format(os.environ["HOME"])
    self.device = device
    self.prediction_distance = prediction_distance
    self.x = torch.from_numpy(np.reshape(np.load(self.path),(-1,10)))[1000000:,:]
    self.augmentation = augmentation

    if split > 0.5:
      self.x = self.x[:int(self.x.shape[0]*split)]
    else:
      self.x = self.x[-int(self.x.shape[0]*split):]
    ss = StandardScaler()
    self.current_index = 0
    self.x_index_address = {}
    self.y_index_address = {}
    self.x = ss.fit_transform(self.x)
    self.window = window_size
    self.generate_percent_labels()
    self.n_features = self.x.shape[1]
    self.n_classes = len(np.unique(self.y))
    
    if samples_per_class != None:
      self.samples_per_class = [samples_per_class]* self.n_classes
    else:
      self.samples_per_class = [0]* self.n_classes
    
    for i in range(0,self.x.shape[0]-window_size,10):
      self.add_to_dataset(self.x[i:i+500],self.y[i:i+500])
    
    for i in range(self.n_classes):
      print(np.count_nonzero(self.y == i))
    print(len(np.unique(self.y)))
    print(self.samples_per_class)
    self.n_samples = self.current_index
  def generate_labels(self):

    idx = 3
    self.y = []
    for i in range(self.n_samples- self.prediction_distance):
        if self.x[i+self.prediction_distance,idx] > self.x[i,idx]:
          self.y.append(1)
        else:
          self.y.append(0)
    print(self.x.shape[0])

  def add_to_dataset(self,x,y):
    self.x_index_address[self.current_index] = torch.from_numpy(np.swapaxes(x,0,1)).float()
    self.y_index_address[self.current_index] = torch.tensor(y[-1]).long()
    self.samples_per_class[int(y[0])] += 1
    self.current_index += 1
    
    ##DEBUG

  def generate_percent_labels(self, classes = [0.01,0.1,0.25,0.5,0.75,0.9,0.99,1]):
    abs_change = np.diff(self.x[:,3])
    self.class_labels = {}
    per_change = np.divide(abs_change , self.x[1:,3])
    i_last = 0 
    self.y = np.zeros(self.x.shape[0])
    for c,i in enumerate(classes):
      self.class_labels[c] = "{}-{}".format(i_last,i)
      print(np.nanquantile(per_change, i))
      hold = np.where( (per_change > np.nanquantile(per_change,i_last)) & (per_change < np.nanquantile(per_change,i)))
      i_last = i 
      self.y[hold] = c
    
    print(len(self.y))
    print(self.x.shape[0])



  def set_window_size(self, window_size):
    self.window = window_size
  def __getitem__(self, index):
    x = self.x_index_address[index].cuda(device = self.device)
    y = self.y_index_address[index].cuda(device = self.device)
    if self.augmentation:
      for func in self.augmentation:
        x, y = func(x,y)
    return x, y
  
  def get_n_classes(self):
    return self.n_classes
  def get_n_samples(self):
    return self.n_samples
  def get_n_features(self):
    return self.n_features
  def __len__(self):
    return self.n_samples
"""

class BTC(Dataset):
  def __init__(self,window_size,split,augmentation,device,_set,window_step = 1):
    self.x = np.load("/home/cmackinnon/scripts/datasets/BTC/BTC_x.npy")
    self.y = np.load("/home/cmackinnon/scripts/datasets/BTC/BTC_y.npy")
    self.n_classes = 1
    print(self.x.shape,split)
    if _set == "train":
        self.x,self.y = self.x[:int(split*self.x.shape[0])], self.y[:int(split*self.y.shape[0])]
    else:
        self.x,self.y = self.x[int((1-split)*self.x.shape[0]):], self.y[int((1-split)*self.y.shape[0]):]
    self.n_samples = int((self.x.shape[0] - window_size) / window_step)
    self.window_size = window_size
    self.augmentation = augmentation
    self.x ,self.y = torch.from_numpy(self.x).cuda(device), torch.from_numpy(self.y).cuda(device)
    print(self.x.shape)
    self.window_step = window_step
    self.x = torch.swapaxes(self.x,0,1).float()
    self.n_features = self.x.shape[0]
  def __getitem__(self, index):
    index *=self.window_step
    x = self.x[:,index:self.window_size+index]
    y = self.y[index+self.window_size]
    if self.augmentation:
      for func in self.augmentation:
        x, y = func(x,y)
    return x, y
  
  def get_n_classes(self):
    return self.n_classes
  def get_n_samples(self):
    return self.n_samples
  def get_n_features(self):
    return self.n_features
  def __len__(self):
    return self.n_samples

class Train(BTC):

  def __init__(self, window_size = 500, split = 0.95 , augmentation = False, device = None,window_step = 10): 
    super().__init__(window_size, split = split, augmentation = augmentation, device =device, _set = "train",window_step = window_step)

class Test(BTC):
  def __init__(self, window_size = 500, split = 0.05, augmentation = False,device = None): 
    super().__init__(window_size, split = split, augmentation = augmentation , device =device, _set = "test")





import torch
import numpy as np
import math
import torch.nn as nn
import torch.nn.functional as F
ACTIVATION = {
  "gelu" : lambda C, L: nn.GELU(),
  "relu" : lambda  C, L: nn.ReLU(),
  "none" : lambda  C, L: Identity(),
  "tanh" : lambda  C, L: nn.Tanh(),
  "mha" : lambda  C, L: MHA(C)
}
NORMALISATION = {
  "batch_norm" : lambda C, l : nn.BatchNorm1d(C),  
  "layer_norm" : lambda C, l : nn.LayerNorm([C,l]),
  "none" : lambda C, l : Identity()

}



OPS = {
  "gelu" : lambda C,kernel, stride,dil, affine: nn.GELU(),
  "relu" : lambda : nn.ReLU(),
  "batch_norm" : lambda C : nn.BatchNorm1d(C),
  "skip_connect" : lambda C : Identity(),
  "resample_channels" : lambda c_in,c_out: ResampleChannels(c_in,c_out),
  "downsample_resolution" : lambda c_in,stride: DownsampleResolution(c_in,stride),
  "point_conv_1" :  lambda C: PointConv(C, stride = 1,dilation = 1,padding = 1*(3 - 1)//2, affine = True),
  #DEPTHWISE CONVOLUTIONS
  "depth_conv_3_1" :  lambda C: DepthConv(C,3, stride = 1,dilation = 1,padding = 1*(3 - 1)//2, affine = True),
  "depth_conv_5_1" :  lambda C: DepthConv(C,5, stride = 1,dilation = 1,padding = 1*(5 - 1)//2, affine = True),
  "depth_conv_7_1" :  lambda C: DepthConv(C,7, stride = 1,dilation = 1,padding = 1*(7 - 1)//2, affine = True),
  "depth_conv_3_2" :  lambda C: DepthConv(C,3, stride = 1,dilation = 2,padding = 2*(3 - 1)//2, affine = True),
  "depth_conv_5_2" :  lambda C: DepthConv(C,5, stride = 1,dilation = 2,padding = 2*(5 - 1)//2, affine = True),
  "depth_conv_7_2" :  lambda C: DepthConv(C,7, stride = 1,dilation = 2,padding = 2*(7 - 1)//2, affine = True),
  "depth_conv_3_4" :  lambda C: DepthConv(C,3, stride = 1,dilation = 4,padding = 4*(3 - 1)//2, affine = True),
  "depth_conv_5_4" :  lambda C: DepthConv(C,5, stride = 1,dilation = 4,padding = 4*(5 - 1)//2, affine = True),
  "depth_conv_7_4" :  lambda C: DepthConv(C,7, stride = 1,dilation = 4,padding = 4*(7 - 1)//2, affine = True),
  "depth_conv_3_8" :  lambda C: DepthConv(C,3, stride = 1,dilation = 8,padding = 8*(3 - 1)//2, affine = True),
  "depth_conv_5_8" :  lambda C: DepthConv(C,5, stride = 1,dilation = 8,padding = 8*(5 - 1)//2, affine = True),
  "depth_conv_7_8" :  lambda C: DepthConv(C,7, stride = 1,dilation = 8,padding = 8*(7 - 1)//2, affine = True),
  "depth_conv_3_16" :  lambda C: DepthConv(C,3, stride = 1,dilation = 16,padding = 16*(3 - 1)//2, affine = True),
  "depth_conv_5_16" :  lambda C: DepthConv(C,5, stride = 1,dilation = 16,padding = 16*(5 - 1)//2, affine = True),
  "depth_conv_7_16" :  lambda C: DepthConv(C,7, stride = 1,dilation = 16,padding = 16*(7 - 1)//2, affine = True),
  "depth_conv_3_32" :  lambda C: DepthConv(C,3, stride = 1,dilation = 32,padding = 32*(3 - 1)//2, affine = True),
  "depth_conv_5_32" :  lambda C: DepthConv(C,5, stride = 1,dilation = 32,padding = 32*(5 - 1)//2, affine = True),
  "depth_conv_7_32" :  lambda C: DepthConv(C,7, stride = 1,dilation = 32,padding = 32*(7 - 1)//2, affine = True),
  "depth_conv_15_1" :  lambda C: DepthConv(C,15, stride = 1,dilation = 1,padding = 1*(15 - 1)//2, affine = True),
  "depth_conv_31_1" :  lambda C: DepthConv(C,31, stride = 1,dilation = 1,padding = 1*(31 - 1)//2, affine = True),
  "depth_conv_63_1" :  lambda C: DepthConv(C,63, stride = 1,dilation = 1,padding = 1*(63 - 1)//2, affine = True),
  "depth_conv_15_2" :  lambda C: DepthConv(C,15, stride = 1,dilation = 2,padding = 2*(15 - 1)//2, affine = True),
  "depth_conv_31_2" :  lambda C: DepthConv(C,31, stride = 1,dilation = 2,padding = 2*(31 - 1)//2, affine = True),
  "depth_conv_63_2" :  lambda C: DepthConv(C,63, stride = 1,dilation = 2,padding = 2*(63 - 1)//2, affine = True),
  "depth_conv_15_4" :  lambda C: DepthConv(C,15, stride = 1,dilation = 4,padding = 4*(15 - 1)//2, affine = True),
  "depth_conv_31_4" :  lambda C: DepthConv(C,31, stride = 1,dilation = 4,padding = 4*(31 - 1)//2, affine = True),
  "depth_conv_63_4" :  lambda C: DepthConv(C,63, stride = 1,dilation = 4,padding = 4*(63 - 1)//2, affine = True),
  "depth_conv_15_8" :  lambda C: DepthConv(C,15, stride = 1,dilation = 8,padding = 8*(15 - 1)//2, affine = True),
  "depth_conv_31_8" :  lambda C: DepthConv(C,31, stride = 1,dilation = 8,padding = 8*(31 - 1)//2, affine = True),
  "depth_conv_63_8" :  lambda C: DepthConv(C,63, stride = 1,dilation = 8,padding = 8*(63 - 1)//2, affine = True),
  "depth_conv_15_16" :  lambda C: DepthConv(C,15, stride = 1,dilation = 16,padding =16*(15 - 1)//2, affine = True),
  "depth_conv_63_32" :  lambda C: DepthConv(C,63, stride = 1,dilation = 32,padding =32*(63 - 1)//2, affine = True),
  #AVERAGE POOLING OPERATIONS
  "avg_pool_3_1" :  lambda C: nn.AvgPool1d(3,stride = 1, padding = 1*(3 - 1)//2,count_include_pad=False),
  "avg_pool_5_1" :  lambda C: nn.AvgPool1d(5,stride = 1,padding = 2,count_include_pad=False),
  "avg_pool_7_1" :  lambda C: nn.AvgPool1d(7,stride = 1,padding = 3,count_include_pad=False),
  "avg_pool_15_1" :  lambda C: nn.AvgPool1d(15,stride = 1,padding = 7,count_include_pad=False),
  "avg_pool_31_1" :  lambda C: nn.AvgPool1d(31,stride = 1,padding = 15,count_include_pad=False),
  "avg_pool_63_1" :  lambda C: nn.AvgPool1d(63,stride = 1,padding = 31,count_include_pad=False),
  #MAXPOOLING OPERATIONS
  "max_pool_3_1" :  lambda C: nn.MaxPool1d(3,stride = 1 , dilation = 1,padding = 1),
  "max_pool_5_1" :  lambda C: nn.MaxPool1d(5,stride = 1 , dilation = 1,padding = 2),
  "max_pool_7_1" :  lambda C: nn.MaxPool1d(7,stride = 1 , dilation = 1,padding = 3),
  "max_pool_15_1" :  lambda C: nn.MaxPool1d(15,stride = 1 , dilation = 1,padding = 7),
  "max_pool_31_1" :  lambda C: nn.MaxPool1d(31,stride = 1 , dilation = 1,padding = 15),
  "max_pool_63_1" :  lambda C: nn.MaxPool1d(63,stride = 1 , dilation = 1,padding = 31),
  "max_pool_3_2" :  lambda C: nn.MaxPool1d(3,stride = 1 , dilation = 2,padding = 2*(3 - 1)//2),
  "max_pool_5_2" :  lambda C: nn.MaxPool1d(5,stride = 1 , dilation = 2,padding = 2*(5 - 1)//2),
  "max_pool_7_2" :  lambda C: nn.MaxPool1d(7,stride = 1 , dilation = 2,padding = 2*(7 - 1)//2),
  "max_pool_3_4" :  lambda C: nn.MaxPool1d(3,stride = 1 , dilation = 4,padding = 4*(3 - 1)//2),
  "max_pool_5_4" :  lambda C: nn.MaxPool1d(5,stride = 1 , dilation = 4,padding = 4*(5 - 1)//2),
  "max_pool_7_4" :  lambda C: nn.MaxPool1d(7,stride = 1 , dilation = 4,padding = 4*(7 - 1)//2),
  "max_pool_3_8" :  lambda C: nn.MaxPool1d(3,stride = 1 , dilation = 8,padding = 8*(3 - 1)//2),
  "max_pool_5_8" :  lambda C: nn.MaxPool1d(5,stride = 1 , dilation = 8,padding = 8*(5 - 1)//2),
  "max_pool_7_8" :  lambda C: nn.MaxPool1d(7,stride = 1 , dilation = 8,padding = 8*(7 - 1)//2),
  "max_pool_3_16" :  lambda C: nn.MaxPool1d(3,stride = 1 , dilation = 16,padding = 16*(3 - 1)//2),
  "max_pool_5_16" :  lambda C: nn.MaxPool1d(5,stride = 1 , dilation = 16,padding = 16*(5 - 1)//2),
  "max_pool_7_16" :  lambda C: nn.MaxPool1d(7,stride = 1 , dilation = 16,padding = 16*(7 - 1)//2),
  "max_pool_3_32" :  lambda C: nn.MaxPool1d(3,stride = 1 , dilation = 32,padding = 32*(3 - 1)//2),
  "max_pool_5_32" :  lambda C: nn.MaxPool1d(5,stride = 1 , dilation = 32,padding = 32*(5 - 1)//2),
  "max_pool_7_32" :  lambda C: nn.MaxPool1d(7,stride = 1 , dilation = 32,padding = 32*(7 - 1)//2),
  "max_pool_15_2" :  lambda C: nn.MaxPool1d(15,stride = 1 , dilation = 2,padding = 2*(15 - 1)//2),
  "max_pool_31_2" :  lambda C: nn.MaxPool1d(31,stride = 1 , dilation = 2,padding = 2*(31 - 1)//2),
  "max_pool_63_2" :  lambda C: nn.MaxPool1d(63,stride = 1 , dilation = 2,padding = 2*(63 - 1)//2),
  "max_pool_15_4" :  lambda C: nn.MaxPool1d(15,stride = 1 , dilation = 4,padding = 4*(15 - 1)//2),
  "max_pool_31_4" :  lambda C: nn.MaxPool1d(31,stride = 1 , dilation = 4,padding = 4*(31 - 1)//2),
  "max_pool_63_4" :  lambda C: nn.MaxPool1d(63,stride = 1 , dilation = 4,padding = 4*(63 - 1)//2),
  "max_pool_15_8" :  lambda C: nn.MaxPool1d(15,stride = 1 , dilation = 8,padding = 8*(15 - 1)//2),
  "max_pool_31_8" :  lambda C: nn.MaxPool1d(31,stride = 1 , dilation = 8,padding = 8*(31 - 1)//2),
  "max_pool_63_8" :  lambda C: nn.MaxPool1d(63,stride = 1 , dilation = 8,padding = 8*(63 - 1)//2),
  "max_pool_15_16" :  lambda C: nn.MaxPool1d(15,stride = 1 , dilation = 16,padding =16*(15 - 1)//2),
  "max_pool_31_16" :  lambda C: nn.MaxPool1d(31,stride = 1 , dilation = 16,padding =16*(31 - 1)//2),
  "max_pool_63_16" :  lambda C: nn.MaxPool1d(63,stride = 1 , dilation = 16,padding =16*(63 - 1)//2),
  "max_pool_15_32" :  lambda C: nn.MaxPool1d(15,stride = 1 , dilation = 32,padding =32*(15 - 1)//2),
  "max_pool_31_32" :  lambda C: nn.MaxPool1d(31,stride = 1 , dilation = 32,padding =32*(31 - 1)//2),
  "max_pool_63_32" :  lambda C: nn.MaxPool1d(63,stride = 1 , dilation = 32,padding =32*(63 - 1)//2),
  
  
  }
class MHA(nn.Module):
  def __init__(self,C):
    super(MHA,self).__init__()
    self.act = nn.MultiheadAttention(C, C, batch_first = True)
  def forward(self,x):
    x = x.swapaxes(1,2)
    return self.act(x,x,x,need_weights=False)[0].swapaxes(1,2)
  

class Identity(nn.Module):

  def __init__(self):
    super(Identity, self).__init__()

  def forward(self, x):
    return x


class Zero(nn.Module):

  def __init__(self, stride):
    super(Zero, self).__init__()
    self.stride = stride

  def forward(self, x):
    if self.stride == 1:
      return x.mul(0.)
    return x[:,:,::self.stride,::self.stride,::self.stride].mul(0.)

class DownsampleResolution(nn.Module):
  """
  This operation uses a depthwise convolution to reduce the signal resolution when needed.
  Given a stride of 2 its a kernel of 2 etc. 
  """
  def __init__(self, c_in,stride):
    super(DownsampleResolution,self).__init__()
    self.conv1 = nn.Conv3d(c_in,c_in, kernel_size =stride, stride = stride, bias = False, groups = c_in)
  def forward(self,x):
    return self.conv1(x)

class ResampleChannels(nn.Module):
  """
  This operation uses a pointwise convolution to change the number of channels when needed.
  """
  def __init__(self,c_in,c_out):
    super(ResampleChannels, self).__init__()
    self.conv1 = nn.Conv3d(c_in,c_out, kernel_size = 1, stride = 1, bias = False)
  def forward(self,x):
    return self.conv1(x)

class GELU(nn.Module):
  def __init__(self):
    super(GELU,self).__init__()
    self.act = nn.GELU()
  def forward(self,x):
    return self.act(x)

class SEMIX(nn.Module):
  def __init__(self, C_in,C_out,r =2 ,stride =1,affine = True ):
    super(SEMIX,self).__init__()
    #print("Building Squeeze Excite with input {} and output: {}".format(C_in,C_out))
    self.GP = nn.AdaptiveAvgPool3d(1)
    self.fc1 = nn.Linear(C_in, C_in//2, bias = False)
    self.act = nn.GELU()
    self.fc2 = nn.Linear(C_in//2, C_out ,bias = False)
    self.sig = nn.Sigmoid()
    self.stride = stride
  def forward(self,x1,x2):
    #Squeeze
    y = self.GP(x2).squeeze()# [Batch,C]
    #torch.mean(x,axis = 2)  
    y = self.fc1(y)
    y = self.act(y)
    y = self.sig(y).unsqueeze(dim = 2)
    return x1* y.expand_as(x1)


class LayerNorm(nn.Module):
  def __init__(self,c,affine):
    super(LayerNorm,self).__init__()
    self.norm = nn.LayerNorm(c, affine=affine)
  def forward(self,x):
    return self.norm(x)

class BatchNorm(nn.Module):
  def __init__(self,c,affine):
    super(BatchNorm,self).__init__()
    self.norm = nn.BatchNorm3d(c, affine=affine)
  def forward(self,x):
    return self.norm(x)

class DepthConv(nn.Module):
  def __init__(self, c, kernel_size, padding,stride,dilation= 1, affine=True):
    super(DepthConv, self).__init__()
    #padding = (kernel_size*stride*dilation)//2
    self.conv = nn.Conv3d(c,c, kernel_size=kernel_size, stride=stride, padding="same", groups=c, bias=False)

  def forward(self, x):
    x =self.conv(x)
    return x

class PointConv(nn.Module):
  def __init__(self, C, stride,affine=True,dilation = 1,padding = 0 ):
    super(PointConv, self).__init__()
    if stride == 2:
      self.conv = nn.Conv3d(C,C*2, kernel_size=1, dilation = 1,stride=stride, bias=False)
    else:
      self.conv = nn.Conv3d(C,C, kernel_size=1, stride=stride,dilation = 1 , bias=False)

  def forward(self, x):
    x =self.conv(x)
    return x



import numpy as np 
import time
import os 
import matplotlib.pyplot as plt
from HPO.utils.model import NetworkMain
from HPO.utils.DARTS_utils import config_space_2_DARTS
from HPO.utils.FCN import FCN 
import pandas as pd
import torch
from HPO.data.btc_dataset import Train , Test
import torch.nn as nn
from torch import Tensor
from torch.utils.data import DataLoader, SubsetRandomSampler
import random
import HPO.utils.augmentation as aug
from HPO.utils.worker_train import train_model, collate_fn_padd, train_model_bt, collate_fn_padd_x, train_model_aug, train_model_multibatch
from HPO.utils.weight_freezing import freeze_FCN, freeze_resnet
from HPO.utils.ResNet1d import resnet18
from HPO.utils.files import save_obj
from queue import Empty
from sklearn.model_selection import StratifiedKFold as KFold
from collections import namedtuple
from HPO.utils.worker_score import Evaluator 
from HPO.utils.worker_utils import LivePlot
Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')


def compute( ID = None, configs=None , gpus=None , res = None  , config = None):
  device = None
  print("Starting process: {}".format(ID))
  while not configs.empty():
    try:
      if device == None:
        device = gpus.get(timeout = 10)
      config = configs.get(timeout = 10)
    except Empty:
      if device != None:
        gpus.put(device)
      
    except:
      torch.cuda.empty_cache()
      if device != None:
        gpus.put(device)
      return
    if config != None:
      print("Got Configuration!")

    if device != None:
      print("Starting config with device: {}".format(device))
      complete = False
      crashes = 0
      acc , rec =  _compute(hyperparameter = config , cuda_device = device)
      while not complete:
        try:
          
          complete = True
        except:
          crashes +=1
          print("Model crash: {} ".format(crashes))
          time.sleep(60)
          if crashes == 2:
            print("Final Crash giving score of zero")
            acc , rec = 0 , 0 
            complete = True
      res.put([config , acc , rec ]) 

  torch.cuda.empty_cache()
   
def _compute(hyperparameter,budget = 1, in_model = None , train_path = None,  test_path = None, cuda_device = None,plot_queue = None, model_id = None, binary = True):
  ### Configuration 
  THRESHOLD = 0.4 #Cut off for classification
  batch_size = 256
  if cuda_device == None:
     cuda_device = 0# torch.cuda.current_device()
  
  ##Set up augmentations##
  jitter = aug.Jitter(device = cuda_device,sigma = 0.1)
  crop = aug.Crop(device = cuda_device)
  scaling = aug.Scaling(device = cuda_device)
  window_warp = aug.WindowWarp(device = cuda_device)
  cut_out = aug.CutOut(device = cuda_device)
  augmentations = [jitter,crop,scaling, window_warp, cut_out]

  dataset_train = Train(device = cuda_device)
  dataset_test = Test()
  #dataset_test_full = Test_TEPS()
  torch.cuda.set_device(cuda_device)

  print("Cuda Device Value: ", cuda_device)
  gen = config_space_2_DARTS(hyperparameter,reduction = True)
  print(gen)

  gen = Genotype(normal=[('avg_pool_3x3', 1), ('avg_pool_3x3', 0), ('avg_pool_3x3', 1), ('dil_conv_3x3', 0), ('avg_pool_3x3', 1), ('dil_conv_5x5', 0), ('sep_conv_5x5', 4), ('dil_conv_5x5', 2)],normal_concat=range(2, 6), reduce=[('max_pool_3x3', 0), ('skip_connect', 1), ('avg_pool_3x3', 0), ('dil_conv_3x3', 1), ('dil_conv_3x3', 2), ('sep_conv_5x5', 1), ('sep_conv_5x5', 2), ('dil_conv_5x5', 1)], reduce_concat=range(2, 6))


  n_classes = dataset_train.get_n_classes()
  inner = 4
  kfold = KFold(n_splits = inner, shuffle = True)
  evaluator = Evaluator(batch_size, n_classes,cuda_device) 
  multibatch = False
  for fold in range(budget):
      print('---Fold No.--{}----------------------'.format(fold))
      torch.cuda.empty_cache()
      if multibatch:
        batches = [2,4,8,16]
        trainloader = {}
        for i in batches:
          trainloader[i] = torch.utils.data.DataLoader(
                              dataset_train,collate_fn = collate_fn_padd,shuffle = True,
                              batch_size=i, drop_last = True)
      else:
        trainloader = torch.utils.data.DataLoader(
                          dataset_train,collate_fn = collate_fn_padd,shuffle = True,
                          batch_size=batch_size, drop_last = True)
      testloader = torch.utils.data.DataLoader(
                          dataset_test,collate_fn = collate_fn_padd,
                          batch_size=batch_size,drop_last = True)

      #testloader_full = torch.utils.data.DataLoader(
      #                    dataset_test_full,collate_fn = collate_fn_padd,
      #                   batch_size=batch_size,drop_last = True)
      ### Build Model
      model = NetworkMain(dataset_train.get_n_features(),hyperparameter["channels"],num_classes= dataset_train.get_n_classes() , layers = hyperparameter["layers"], auxiliary = False,drop_prob = hyperparameter["p"], genotype = gen, binary = binary)
      model = model.cuda(device = cuda_device)
      """
      ### Train the model
      """
      #pretrain_naswot = evaluator.score_naswot(model,testloader)
      if multibatch:
        train_model_multibatch(model , hyperparameter, trainloader , hyperparameter["epochs"], batch_size , cuda_device, augment_num = 1, graph = plot_queue, binary = binary) 
      else:  
        train_model(model , hyperparameter, trainloader , hyperparameter["epochs"], batch_size , cuda_device, graph = plot_queue, binary = binary) 
      """
      ### Test the model
      """
      evaluator.forward_pass(model, testloader,binary)
      evaluator.predictions(model_is_binary = binary , THRESHOLD = THRESHOLD)

      ### Get Metrics
      total = evaluator.T()
      acc  =  evaluator.T_ACC()
      recall = evaluator.TPR(1)
      recall_total = evaluator.P(1)
      print("Supervised Loss: {} -- Unsupvised Loss: {}".format(sup,unsup))
      #print("Supervised Loss(10): {} -- Unsupervised Loss(10): {}".format(sup10,unsup10))
      #print("NASWOT PRE: {} -- NASWOT POST: {}".format(pretrain_naswot,posttrain_naswot))
      print("Accuracy: ", "%.4f" % ((acc)*100), "%")
      print("Recall: ", "%.4f" % ((recall)*100), "%")

      ### Save Model
      def save_model(model, hyperparameter):
        model_zoo = "{}/scripts/model_zoo/".format(os.environ["HOME"])
        torch.save(model.state_dict() , model_zoo+"-Acc-{}-Rec-{}".format(acc, recall))
        save_obj( hyperparameter , model_zoo+"hps/"+"-Acc-{}-Rec-{}".format(acc , recall) )


  print("Final Scores -- ACC: {} -- REC: {}".format(acc, recall))
  return acc, recall, sup.item(), unsup.item()#, sup10.item(),unsup10.item()

if __name__ == "__main__":
  import multiprocessing
  hyperparameter = {'channels': 128, 'crop': 0.8435749854867439, 'crop_rate': 1.0789375913465558, 'cut_mix': 0.3606216485561041, 'cut_mix_rate': 0.006043016512336742, 'cut_out': 0.3759927924492086, 'cut_out_rate': 0.665600887946834, 'epochs': 80, 'jitter': 0.2565783789758802, 'jitter_rate': 1.4762754789497605, 'layers': 1, 'lr': 0.0016466283692574232, 'mix_up': 0.609992555365736, 'mix_up_rate': 0.36211335694823126, 'normal_index_0_0': 0, 'normal_index_0_1': 1, 'normal_index_1_0': 0, 'normal_index_1_1': 2, 'normal_index_2_0': 1, 'normal_index_2_1': 3, 'normal_index_3_0': 3, 'normal_index_3_1': 3, 'normal_node_0_0': 'avg_pool_3x3', 'normal_node_0_1': 'sep_conv_5x5', 'normal_node_1_0': 'dil_conv_3x3', 'normal_node_1_1': 'skip_connect', 'normal_node_2_0': 'max_pool_3x3', 'normal_node_2_1': 'avg_pool_3x3', 'normal_node_3_0': 'avg_pool_3x3', 'normal_node_3_1': 'sep_conv_5x5', 'p': 0.14391084865105958, 'reduction_index_0_0': 0, 'reduction_index_0_1': 1, 'reduction_index_1_0': 0, 'reduction_index_1_1': 0, 'reduction_index_2_0': 2, 'reduction_index_2_1': 2, 'reduction_index_3_0': 2, 'reduction_index_3_1': 3, 'reduction_node_0_0': 'sep_conv_3x3', 'reduction_node_0_1': 'sep_conv_7x7', 'reduction_node_1_0': 'dil_conv_3x3', 'reduction_node_1_1': 'skip_connect', 'reduction_node_2_0': 'skip_connect', 'reduction_node_2_1': 'avg_pool_3x3', 'reduction_node_3_0': 'dil_conv_3x3', 'reduction_node_3_1': 'sep_conv_7x7', 'scaling': 0.4499911429155192, 'scaling_rate': 1.0704794966200513, 'window_warp_num': 5, 'window_warp_rate': 1.8569110504223219}

  #0.8125,0.6896551724137931
  hyperparameter = {'normal_index_0_0': 1, 'normal_index_0_1': 0, 'normal_index_1_0': 1, 'normal_index_1_1': 0, 'normal_index_2_0': 3, 'normal_index_2_1': 2, 'normal_index_3_0': 0, 'normal_index_3_1': 3, 'normal_node_0_0': 'skip_connect', 'normal_node_0_1': 'dil_conv_5x5', 'normal_node_1_0': 'max_pool_3x3', 'normal_node_1_1': 'max_pool_3x3', 'normal_node_2_0': 'sep_conv_7x7', 'normal_node_2_1': 'none', 'normal_node_3_0': 'avg_pool_3x3', 'normal_node_3_1': 'sep_conv_5x5', 'reduction_index_0_0': 1, 'reduction_index_0_1': 0, 'reduction_index_1_0': 0, 'reduction_index_1_1': 2, 'reduction_index_2_0': 1, 'reduction_index_2_1': 3, 'reduction_index_3_0': 2, 'reduction_index_3_1': 4, 'reduction_node_0_0': 'none', 'reduction_node_0_1': 'avg_pool_3x3', 'reduction_node_1_0': 'max_pool_3x3', 'reduction_node_1_1': 'none', 'reduction_node_2_0': 'max_pool_3x3', 'reduction_node_2_1': 'sep_conv_5x5', 'reduction_node_3_0': 'dil_conv_5x5', 'reduction_node_3_1': 'avg_pool_3x3', 'batch_size': 2, 'channels': 27, 'jitter': 0.1241258424762939, 'jitter_rate': 0.5439942968995378, 'mix_up': 0.19412584247629389, 'mix_up_rate': 0.5439942968995378, 'cut_mix': 0.19412584247629389, 'cut_mix_rate': 0.5439942968995378, 'cut_out': 0.0941258424762939, 'cut_out_rate': 0.7439942968995378, 'crop': 0.19412584247629389, 'crop_rate': 0.5439942968995378, 'scaling': 0.001317169415702424, 'scaling_rate': 0.4353430973459786, 'window_warp_num': 3, 'window_warp_rate': 1.4001548161604196, 'lr': 0.005170869707739693, 'p': 0.00296905723528657, 'epochs': 70, 'layers': 3}

  hyperparameter = {'normal_index_0_0': 0, 'normal_index_0_1': 0, 'normal_index_1_0': 1, 'normal_index_1_1': 2, 'normal_index_2_0': 0, 'normal_index_2_1': 3, 'normal_index_3_0': 2, 'normal_index_3_1': 2, 'normal_node_0_0': 'avg_pool_3x3', 'normal_node_0_1': 'none', 'normal_node_1_0': 'max_pool_3x3', 'normal_node_1_1': 'skip_connect', 'normal_node_2_0': 'sep_conv_7x7', 'normal_node_2_1': 'dil_conv_3x3', 'normal_node_3_0': 'skip_connect', 'normal_node_3_1': 'dil_conv_3x3', 'reduction_index_0_0': 0, 'reduction_index_0_1': 0, 'reduction_index_1_0': 2, 'reduction_index_1_1': 1, 'reduction_index_2_0': 3, 'reduction_index_2_1': 0, 'reduction_index_3_0': 1, 'reduction_index_3_1': 4, 'reduction_node_0_0': 'sep_conv_3x3', 'reduction_node_0_1': 'sep_conv_5x5', 'reduction_node_1_0': 'sep_conv_7x7', 'reduction_node_1_1': 'skip_connect', 'reduction_node_2_0': 'max_pool_3x3', 'reduction_node_2_1': 'skip_connect', 'reduction_node_3_0': 'max_pool_3x3', 'reduction_node_3_1': 'dil_conv_5x5', 'batch_size': 2, 'channels': 27, 'jitter': 0.1241258424762939, 'jitter_rate': 0.5439942968995378, 'mix_up': 0.19412584247629389, 'mix_up_rate': 0.5439942968995378, 'cut_mix': 0.19412584247629389, 'cut_mix_rate': 0.5439942968995378, 'cut_out': 0.0941258424762939, 'cut_out_rate': 0.7439942968995378, 'crop': 0.19412584247629389, 'crop_rate': 0.5439942968995378, 'scaling': 0.001317169415702424, 'scaling_rate': 0.4353430973459786, 'window_warp_num': 3, 'window_warp_rate': 1.4001548161604196, 'lr': 0.005170869707739693, 'p': 0.00296905723528657, 'epochs': 70, 'layers': 3}
  hpo = {'batch_size': 2, 'channels': 64, 'jitter': 0.01241258424762939, 'jitter_rate': 0.5439942968995378, 'mix_up': 0.19412584247629389, 'mix_up_rate': 0.5439942968995378, 'cut_mix': 0.19412584247629389, 'cut_mix_rate': 0.5439942968995378, 'cut_out': 0.0941258424762939, 'cut_out_rate': 0.7439942968995378, 'crop': 0.19412584247629389, 'crop_rate': 0.5439942968995378, 'scaling': 0.001317169415702424, 'scaling_rate': 0.4353430973459786, 'window_warp_num': 3, 'window_warp_rate': 1.4001548161604196, 'lr': 0.0025170869707739693, 'p': 0.00, 'epochs': 20, 'layers': 3}
  #0.8125,0.7931034482758621,"
  hyperparameter = {'normal_index_0_0': 0, 'normal_index_0_1': 0, 'normal_index_1_0': 2, 'normal_index_1_1': 0, 'normal_index_2_0': 0, 'normal_index_2_1': 0, 'normal_index_3_0': 2, 'normal_index_3_1': 2, 'normal_node_0_0': 'avg_pool_3x3', 'normal_node_0_1': 'none', 'normal_node_1_0': 'skip_connect', 'normal_node_1_1': 'max_pool_3x3', 'normal_node_2_0': 'sep_conv_5x5', 'normal_node_2_1': 'none', 'normal_node_3_0': 'avg_pool_3x3', 'normal_node_3_1': 'dil_conv_3x3', 'reduction_index_0_0': 0, 'reduction_index_0_1': 1, 'reduction_index_1_0': 2, 'reduction_index_1_1': 0, 'reduction_index_2_0': 3, 'reduction_index_2_1': 1, 'reduction_index_3_0': 1, 'reduction_index_3_1': 2, 'reduction_node_0_0': 'skip_connect', 'reduction_node_0_1': 'none', 'reduction_node_1_0': 'max_pool_3x3', 'reduction_node_1_1': 'avg_pool_3x3', 'reduction_node_2_0': 'skip_connect', 'reduction_node_2_1': 'sep_conv_5x5', 'reduction_node_3_0': 'sep_conv_5x5', 'reduction_node_3_1': 'sep_conv_3x3'}

  #0.8170731707317073,0.6486486486486487,"
  hyperparameter = {'normal_index_0_0': 0, 'normal_index_0_1': 0, 'normal_index_1_0': 1, 'normal_index_1_1': 1, 'normal_index_2_0': 3, 'normal_index_2_1': 3, 'normal_index_3_0': 2, 'normal_index_3_1': 2, 'normal_node_0_0': 'avg_pool_3x3', 'normal_node_0_1': 'sep_conv_7x7', 'normal_node_1_0': 'sep_conv_5x5', 'normal_node_1_1': 'sep_conv_5x5', 'normal_node_2_0': 'sep_conv_7x7', 'normal_node_2_1': 'avg_pool_3x3', 'normal_node_3_0': 'skip_connect', 'normal_node_3_1': 'sep_conv_5x5', 'reduction_index_0_0': 1, 'reduction_index_0_1': 0, 'reduction_index_1_0': 0, 'reduction_index_1_1': 1, 'reduction_index_2_0': 1, 'reduction_index_2_1': 2, 'reduction_index_3_0': 4, 'reduction_index_3_1': 1, 'reduction_node_0_0': 'sep_conv_3x3', 'reduction_node_0_1': 'dil_conv_5x5', 'reduction_node_1_0': 'none', 'reduction_node_1_1': 'max_pool_3x3', 'reduction_node_2_0': 'dil_conv_3x3', 'reduction_node_2_1': 'avg_pool_3x3', 'reduction_node_3_0': 'none', 'reduction_node_3_1': 'sep_conv_5x5'}
  hyperparameter.update(hpo)
  queue = multiprocessing.Queue()
  plotter = LivePlot(queue)
  plot_process = multiprocessing.Process(target=plotter.show,args=())
  plot_process.start()
  _compute(hyperparameter, binary = False,plot_queue = queue)
  exit()

